# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import datetime
import logging
import os
import re
import time
import uuid
import warnings
from functools import cached_property

import numpy as np

from climetlab.core.order import build_remapping  # noqa:F401
from climetlab.utils import progress_bar
from climetlab.utils.config import LoadersConfig
from climetlab.utils.humanize import bytes, seconds


class DatasetName:
    def __init__(
        self,
        name,
        resolution=None,
        start_date=None,
        end_date=None,
        frequency=None,
    ):
        self.name = name
        self.parsed = self._parse(name)

        self.messages = []

        self.check_parsed()
        self.check_resolution(resolution)
        self.check_frequency(frequency)
        self.check_start_date(start_date)
        self.check_end_date(end_date)

        if self.messages:
            self.messages.append(
                f"{self} is parsed as :"
                + "/".join(f"{k}={v}" for k, v in self.parsed.items())
            )

    @property
    def is_valid(self):
        return not self.messages

    @property
    def error_message(self):
        out = " And ".join(self.messages)
        if out:
            out = out[0].upper() + out[1:]
        return out

    def raise_if_not_valid(self, print=print):
        if not self.is_valid:
            for m in self.messages:
                print(m)
            raise ValueError(self.error_message)

    def _parse(self, name):
        pattern = r"^(\w+)-(\w+)-(\w+)-(\w+)-(\w\w\w\w)-(\w+)-(\w+)-([\d\-]+)-(\d+h)-v(\d+)-?(.*)$"
        match = re.match(pattern, name)

        parsed = {}
        if match:
            keys = [
                "use_case",
                "class_",
                "type_",
                "stream",
                "expver",
                "source",
                "resolution",
                "period",
                "frequency",
                "version",
                "additional",
            ]
            parsed = {k: v for k, v in zip(keys, match.groups())}

            period = parsed["period"].split("-")
            assert len(period) in (1, 2), (name, period)
            parsed["start_date"] = period[0]
            if len(period) == 1:
                parsed["end_date"] = period[0]
            if len(period) == 2:
                parsed["end_date"] = period[1]

        return parsed

    def __str__(self):
        return self.name

    def check_parsed(self):
        if not self.parsed:
            self.messages.append(
                (
                    f"the dataset name {self} does not follow naming convention. "
                    "See here for details: "
                    "https://confluence.ecmwf.int/display/DWF/Datasets+available+as+zarr"
                )
            )

    def check_resolution(self, resolution):
        if (
            self.parsed.get("resolution")
            and self.parsed["resolution"][0] not in "0123456789on"
        ):
            self.messages.append(
                (
                    f"the resolution {self.parsed['resolution'] } should start "
                    f"with a number or 'o' or 'n' in the dataset name {self}."
                )
            )

        if resolution is None:
            return
        resolution_str = str(resolution).replace(".", "p").lower()
        self._check_missing("resolution", resolution_str)
        self._check_mismatch("resolution", resolution_str)

    def check_frequency(self, frequency):
        if frequency is None:
            return
        frequency_str = f"{frequency}h"
        self._check_missing("frequency", frequency_str)
        self._check_mismatch("frequency", frequency_str)

    def check_start_date(self, start_date):
        if start_date is None:
            return
        start_date_str = str(start_date.year)
        self._check_missing("first date", start_date_str)
        self._check_mismatch("start_date", start_date_str)

    def check_end_date(self, end_date):
        if end_date is None:
            return
        end_date_str = str(end_date.year)
        self._check_missing("end_date", end_date_str)
        self._check_mismatch("end_date", end_date_str)

    def _check_missing(self, key, value):
        if value not in self.name:
            self.messages.append(
                (f"the {key} is {value}, but is missing in {self.name}.")
            )

    def _check_mismatch(self, key, value):
        if self.parsed.get(key) and self.parsed[key] != value:
            self.messages.append(
                (f"the {key} is {value}, but is {self.parsed[key]} in {self.name}.")
            )


LOG = logging.getLogger(__name__)

VERSION = "0.12"


def get_versions():
    dic = {}

    import climetlab

    dic["climetlab"] = climetlab.__version__

    import earthkit.meteo

    dic["earthkit.meteo"] = earthkit.meteo.__version__

    return dic


def check_data_values(arr, *, name: str, log=[]):
    min, max = arr.min(), arr.max()
    assert not (np.isnan(arr).any()), (name, min, max, *log)

    if min == 9999.0:
        warnings.warn(f"Min value 9999 for {name}")
    if max == 9999.0:
        warnings.warn(f"Max value 9999 for {name}")

    if name == ["lsm", "insolation"]:  # 0. to 1.
        assert max <= 1, (name, min, max, *log)
        assert min >= 0, (name, min, max, *log)

    if name == "2t":  # surface temp between -100 celcius and +100 celcius
        assert max <= 373.15, (name, min, max, *log)
        assert min >= 173.15, (name, min, max, *log)


def check_stats(minimum, maximum, mean, msg, **kwargs):
    tolerance = (abs(minimum) + abs(maximum)) * 0.01
    if (mean - minimum < -tolerance) or (mean - minimum < -tolerance):
        raise ValueError(
            f"Mean is not in min/max interval{msg} : we should have {minimum} <= {mean} <= {maximum}"
        )


def _prepare_serialisation(o):
    if isinstance(o, dict):
        dic = {}
        for k, v in o.items():
            v = _prepare_serialisation(v)
            if k == "order_by":
                # zarr attributes are saved with sort_keys=True
                # and ordered dict are reordered.
                # This is a problem for "order_by"
                # We ensure here that the order_by key contains
                # a list of dict
                v = [{kk: vv} for kk, vv in v.items()]
            dic[k] = v
        return dic

    if isinstance(o, (list, tuple)):
        return [_prepare_serialisation(v) for v in o]

    if o in (None, True, False):
        return o

    if isinstance(o, (str, int, float)):
        return o

    if isinstance(o, (datetime.date, datetime.datetime)):
        return o.isoformat()

    return str(o)


class OffsetView:
    def __init__(self, large_array, *, offset, axis, shape):
        """
        A view on a portion of the large_array.
        'axis' is the axis along which the offset applies.
        'shape' is the shape of the view.
        """
        self.large_array = large_array
        self.offset = offset
        self.axis = axis
        self.shape = shape

    def __setitem__(self, key, values):
        if isinstance(key, slice):
            # Ensure that the slice covers the entire view along the axis.
            assert key.start is None and key.stop is None, key

            # Create a new key for indexing the large array.
            new_key = tuple(
                slice(self.offset, self.offset + values.shape[i])
                if i == self.axis
                else slice(None)
                for i in range(len(self.shape))
            )
        else:
            # For non-slice keys, adjust the key based on the offset and axis.
            new_key = tuple(
                k + self.offset if i == self.axis else k for i, k in enumerate(key)
            )

        self.large_array[new_key] = values


class CubesFilter:
    def __init__(self, *, loader, parts, **kwargs):
        self.loader = loader

        if parts is None:
            self.parts = None
            return

        if len(parts) == 1:
            part = parts[0]
            if part.lower() in ["all", "*"]:
                self.parts = None
                return

            if "/" in part:
                i_chunk, n_chunks = part.split("/")
                i_chunk, n_chunks = int(i_chunk), int(n_chunks)

                total = len(self.loader.registry.get_flags())
                assert i_chunk > 0, f"Chunk number {i_chunk} must be positive."
                if n_chunks > total:
                    warnings.warn(
                        f"Number of chunks {n_chunks} is larger than the total number of chunks: {total}+1."
                    )

                chunk_size = total / n_chunks
                parts = [
                    x
                    for x in range(total)
                    if x >= (i_chunk - 1) * chunk_size and x < i_chunk * chunk_size
                ]

        parts = [int(_) for _ in parts]
        print(f"Running parts: {parts}")
        if not parts:
            warnings.warn(f"Nothing to do for chunk {i_chunk}.")

        self.parts = parts

    def __call__(self, i):
        if self.parts is None:
            return True
        return i in self.parts


class Loader:
    def __init__(self, *, path, config, print=print, **kwargs):
        self.main_config = LoadersConfig(config)
        self.input_handler = self.main_config.input_handler()
        self.path = path
        self.kwargs = kwargs
        self.print = print
        self.registry = ZarrBuiltRegistry(self.path)

    def load(self, **kwargs):
        import zarr

        self.z = zarr.open(self.path, mode="r+")
        self.registry.add_to_history("loading_data_start", parts=kwargs.get("parts"))

        filter = CubesFilter(loader=self, **kwargs)
        ncubes = self.input_handler.n_cubes
        for icube, cubecreator in enumerate(self.input_handler.iter_cubes()):
            if not filter(icube):
                continue
            if self.registry.get_flag(icube):
                print(f" -> Skipping {icube} total={ncubes} (already done)")
                continue
            self.print(f" -> Processing i={icube=} total={ncubes}")

            cube = cubecreator.to_cube()
            shape = cube.extended_user_shape
            chunks = cube.chunking(self.input_handler.output.chunking)
            axis = self.input_handler.output.append_axis

            slice = self.registry.get_slice_for(icube)

            print(f"Building ZARR '{self.path}'")
            self.print(f"Building ZARR (total shape ={shape}) at {slice}, {chunks=}")

            offset = slice.start
            array = OffsetView(self.z["data"], offset=offset, axis=axis, shape=shape)

            self.load_datacube(cube, array)

            self.registry.set_flag(icube)

        self.registry.add_to_history("loading_data_end", parts=kwargs.get("parts"))
        self.registry.add_provenance(name="provenance_load")

    def load_datacube(self, cube, array):
        start = time.time()
        load = 0
        save = 0

        reading_chunks = None
        total = cube.count(reading_chunks)
        self.print(f"Loading datacube {cube}")
        bar = progress_bar(
            iterable=cube.iterate_cubelets(reading_chunks),
            total=total,
            desc=f"Loading datacube {cube}",
        )
        for i, cubelet in enumerate(bar):
            now = time.time()
            data = cubelet.to_numpy()
            bar.set_description(f"{i}/{total} {str(cubelet)} ({data.shape})")
            self.print(f"Loading datacube {cube}: {i}/{total}")
            load += time.time() - now

            j = cubelet.extended_icoords[1]
            check_data_values(
                data[:],
                name=self._variables_names[j],
                log=[i, j, data.shape, cubelet.extended_icoords],
            )

            now = time.time()
            array[cubelet.extended_icoords] = data
            save += time.time() - now

        now = time.time()
        save += time.time() - now

        print("Written")
        self.print_info()
        print("Written.")

        self.print(
            f"Elapsed: {seconds(time.time() - start)},"
            f" load time: {seconds(load)},"
            f" write time: {seconds(save)}."
        )


def add_zarr_dataset(name, nparray, *, zarr_root, overwrite=True, dtype=np.float32):
    if isinstance(nparray, (tuple, list)):
        nparray = np.array(nparray, dtype=dtype)
    a = zarr_root.create_dataset(
        name,
        shape=nparray.shape,
        dtype=dtype,
        overwrite=overwrite,
    )
    a[...] = nparray[...]
    return a


class ZarrBuiltRegistry:
    name_lengths = "_build_lengths"
    name_flags = "_build_flags"
    lengths = None
    flags = None
    z = None

    def __init__(self, path):
        import zarr

        assert isinstance(path, str), path
        self.zarr_path = path
        self.synchronizer = zarr.ProcessSynchronizer(
            os.path.join(self.zarr_path, "registry.sync")
        )

    def get_slice_for(self, i):
        lengths = self.get_lengths()
        assert i >= 0 and i < len(lengths)

        start = sum(lengths[:i])
        stop = sum(lengths[: (i + 1)])
        return slice(start, stop)

    def get_lengths(self):
        z = self._open_read()
        return list(z[self.name_lengths][:])

    def get_flags(self, **kwargs):
        z = self._open_read(**kwargs)
        print(list(z[self.name_flags][:]))
        return list(z[self.name_flags][:])

    def get_flag(self, i):
        z = self._open_read()
        return z[self.name_flags][i]

    def set_flag(self, i, value=True):
        z = self._open_write()
        z.attrs["latest_write_timestamp"] = datetime.datetime.utcnow().isoformat()
        z[self.name_flags][i] = value

    def _open_read(self, sync=True):
        import zarr

        if sync:
            return zarr.open(self.zarr_path, mode="r", synchronizer=self.synchronizer)
        else:
            return zarr.open(self.zarr_path, mode="r")

    def _open_write(self):
        import zarr

        return zarr.open(self.zarr_path, mode="r+", synchronizer=self.synchronizer)

    def create(self, lengths, overwrite=False):
        z = self._open_write()

        add_zarr_dataset(
            self.name_lengths,
            lengths,
            zarr_root=z,
            dtype="i4",
            overwrite=overwrite,
        )
        add_zarr_dataset(
            self.name_flags,
            len(lengths) * [False],
            zarr_root=z,
            dtype=bool,
            overwrite=overwrite,
        )
        z = None
        self.add_to_history("initialised")

    def reset(self, lengths):
        return self.create(lengths, overwrite=True)

    def add_provenance(self, name):
        from ecml_tools.provenance import gather_provenance_info

        z = self._open_write()
        z.attrs[name] = gather_provenance_info()

    def add_to_history(self, action, **kwargs):
        new = dict(
            action=action,
            timestamp=datetime.datetime.utcnow().isoformat(),
        )
        new.update(kwargs)

        z = self._open_write()
        history = z.attrs.get("history", [])
        history.append(new)
        z.attrs["history"] = history


class ZarrLoader(Loader):
    writer = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.z = None
        self.statistics = []

    @classmethod
    def from_config(cls, *, config, path, **kwargs):
        # config is the path to the config file
        # or a dict with the config
        obj = cls(config=config, path=path, **kwargs)
        return obj

    @classmethod
    def from_zarr(cls, *, config, path, **kwargs):
        import zarr

        assert os.path.exists(path), path
        z = zarr.open(path, mode="r")
        config = z.attrs["_create_yaml_config"]
        # config = yaml.safe_load(z.attrs["_yaml_dump"])["_create_yaml_config"]
        kwargs.get("print", print)("Config loaded from zarr: ", config)
        return cls.from_config(config=config, path=path, **kwargs)

    def iter_loops(self):
        for vars in self.input_handler.iter_loops():
            yield vars

    def _compute_lengths(self, multiply):
        def squeeze_dict(dic):
            keys = list(dic.keys())
            assert len(dic) == 1, keys
            return dic[keys[0]]

        lengths = []
        for i, vars in enumerate(self.iter_loops()):
            lst = squeeze_dict(vars)
            assert isinstance(lst, (tuple, list)), lst
            lengths.append(len(lst))
            print("i vars", i, vars, lengths, lst, f"{multiply=}")

        lengths = [x * multiply for x in lengths]
        return lengths

    @property
    def _variables_names(self):
        return self.main_config.output.order_by[self.main_config.output.statistics]

    def initialise(self):
        """Create empty zarr from self.main_config and self.path"""
        import pandas as pd
        import zarr

        self.print("config loaded ok:")
        print(self.main_config)
        print("-------------------------")

        total_shape = self.input_handler.shape
        self.print(f"total_shape = {total_shape}")
        print("-------------------------")

        grid_points = self.input_handler.grid_points
        print(f"gridpoints size: {[len(i) for i in grid_points]}")
        print("-------------------------")

        dates = self.input_handler.get_datetimes()
        self.print(f"Found {len(dates)} datetimes.")
        print(
            f"Dates: Found {len(dates)} datetimes, in {self.input_handler.n_cubes} cubes: ",
            end="",
        )
        lengths = [str(len(c.get_datetimes())) for c in self.input_handler.iter_cubes()]
        print("+".join(lengths))
        self.print(f"Found {len(dates)} datetimes {'+'.join(lengths)}.")
        print("-------------------------")

        variables_names = self.input_handler.variables

        assert (
            variables_names
            == self.main_config.output.order_by[self.main_config.output.statistics]
        ), (
            f"Requested= {self.main_config.output.order_by[self.main_config.output.statistics]} "
            f"Actual= {variables_names}"
        )

        resolution = self.input_handler.resolution

        chunks = self.input_handler.chunking
        dtype = self.main_config.output.dtype

        self.print(
            f"Creating ZARR '{self.path}', with {total_shape=}, "
            f"{chunks=} and {dtype=}"
        )

        frequency = self.input_handler.frequency
        assert isinstance(frequency, int), frequency

        if not self.kwargs["no_check"]:
            basename, ext = os.path.splitext(os.path.basename(self.path))

            ds_name = DatasetName(
                basename,
                resolution,
                dates[0],
                dates[-1],
                frequency,
            )
            ds_name.raise_if_not_valid(print=self.print)

        metadata = {}
        metadata["uuid"] = str(uuid.uuid4())

        metadata.update(self.main_config.get("add_metadata", {}))

        metadata["_create_yaml_config"] = _prepare_serialisation(self.main_config)

        metadata["description"] = self.main_config.description
        metadata["resolution"] = resolution

        metadata["data_request"] = self.input_handler.data_request

        metadata["order_by"] = self.main_config.output.order_by
        metadata["remapping"] = self.main_config.output.remapping
        metadata["flatten_grid"] = self.main_config.output.flatten_grid
        metadata["ensemble_dimension"] = self.main_config.output.ensemble_dimension

        metadata["variables"] = variables_names
        metadata["version"] = VERSION
        metadata["frequency"] = frequency
        metadata["start_date"] = dates[0].isoformat()
        metadata["end_date"] = dates[-1].isoformat()
        pd_dates_kwargs = dict(
            start=metadata["start_date"],
            end=metadata["end_date"],
            freq=f"{metadata['frequency']}h",
            unit="s",
        )
        pd_dates = pd.date_range(**pd_dates_kwargs)

        def check_dates(input_handler, pd_dates, total_shape):
            for i, loop in enumerate(input_handler.loops):
                print(f"Loop {i}: ", loop._info)
            if pd_dates.size != total_shape[0]:
                raise ValueError(
                    f"Final date size {pd_dates.size} (from {pd_dates[0]} to {pd_dates[-1]}, "
                    f"{frequency=}) does not match data shape {total_shape[0]}. {total_shape=}"
                )
            if pd_dates.size != len(dates):
                raise ValueError(
                    f"Final date size {pd_dates.size} (from {pd_dates[0]} to {pd_dates[-1]}, "
                    f"{frequency=}) does not match data shape {len(dates)} (from {dates[0]} to "
                    f"{dates[-1]}). {pd_dates_kwargs}"
                )

        check_dates(self.input_handler, pd_dates, total_shape)

        metadata.update(self.main_config.get("force_metadata", {}))

        # write data
        self.z = zarr.open(self.path, mode="w")

        self.z.create_dataset("data", shape=total_shape, chunks=chunks, dtype=dtype)

        np_dates = pd_dates.to_numpy()
        self._add_dataset("dates", np_dates, dtype=np_dates.dtype)

        self._add_dataset("latitudes", grid_points[0])
        self._add_dataset("longitudes", grid_points[1])

        self.z = None

        self.update_metadata(**metadata)

        self.registry.create(lengths=lengths)

        self.registry.add_to_history("init finished")

    def update_metadata(self, **kwargs):
        import zarr

        z = zarr.open(self.path, mode="w+")
        for k, v in kwargs.items():
            if isinstance(v, np.datetime64):
                v = v.astype(datetime.datetime)
            if isinstance(v, datetime.date):
                v = v.isoformat()
            z.attrs[k] = v

    def statistics_start_indice(self):
        return self._statistics_subset_indices[0]

    def statistics_end_indice(self):
        return self._statistics_subset_indices[1]

    def _actual_statistics_start(self):
        return self._statistics_subset_indices[2]

    def _actual_statistics_end(self):
        return self._statistics_subset_indices[3]

    @cached_property
    def _statistics_subset_indices(self):
        statistics_start = self.main_config.output.get("statistics_start")
        statistics_end = self.main_config.output.get("statistics_end")
        try:
            from ecml_tools.data import open_dataset
        except ImportError:
            raise Exception("Need to pip install ecml_tools[zarr]")

        if statistics_end is None:
            warnings.warn(
                "No statistics_end specified, using last date of the dataset."
            )
        ds = open_dataset(self.path)
        subset = ds.dates_interval_to_indices(statistics_start, statistics_end)

        return (subset[0], subset[-1], ds.dates[subset[0]], ds.dates[subset[-1]])

    def _add_dataset(self, *args, **kwargs):
        import zarr

        z = self.z
        if z is None:
            z = zarr.open(self.path, mode="r+")

        return add_zarr_dataset(*args, **kwargs, zarr_root=z)

    def print_info(self):
        assert self.z is not None
        try:
            print(self.z["data"].info)
        except Exception as e:
            print(e)
        print("...")
        try:
            print(self.z["data"].info)
        except Exception as e:
            print(e)

    def add_statistics(self, no_write, **kwargs):
        do_write = not no_write

        incomplete = not all(self.registry.get_flags(sync=False))
        if do_write and incomplete:
            raise Exception(
                f"Zarr {self.path} is not fully built, not writing statistics."
            )

        statistics_start = self.main_config.output.get("statistics_start")
        statistics_end = self.main_config.output.get("statistics_end")

        if do_write:
            self.registry.add_to_history(
                "compute_statistics_start",
                start=statistics_start,
                end=statistics_end,
            )

        try:
            from ecml_tools.data import open_dataset
        except ImportError:
            raise Exception("Need to pip install ecml_tools")
        ds = open_dataset(self.path)

        stats = self.compute_statistics(ds, statistics_start, statistics_end)

        print(
            "\n".join(
                (
                    f"{v.rjust(10)}: "
                    f"min/max = {stats['minimum'][j]:.6g} {stats['maximum'][j]:.6g}"
                    "   \t:   "
                    f"mean/stdev = {stats['mean'][j]:.6g} {stats['stdev'][j]:.6g}"
                )
                for j, v in enumerate(ds.variables)
            )
        )

        if do_write:
            for k in [
                "mean",
                "stdev",
                "minimum",
                "maximum",
                "sums",
                "squares",
                "count",
            ]:
                self._add_dataset(k, stats[k])

            self.update_metadata(
                statistics_start_date=self._actual_statistics_start(),
                statistics_end_date=self._actual_statistics_end(),
            )

            self.registry.add_to_history(
                "compute_statistics_end",
                start=statistics_start,
                end=statistics_end,
            )
            self.registry.add_provenance(name="provenance_statistics")

    def compute_statistics(self, ds, statistics_start, statistics_end):
        import zarr

        data = zarr.open(self.path, mode="r")["data"]

        shape = data.shape
        assert shape[0] == len(ds.dates)
        assert shape[1] == len(ds.variables)
        assert ds.variables == self._variables_names

        i_start = self.statistics_start_indice()
        i_end = self.statistics_end_indice()

        i_len = i_end + 1 - i_start

        self.print(
            f"Statistics computed on {i_len}/{shape[0]} samples "
            f"first={ds.dates[i_start]} "
            f"last={ds.dates[i_end]}"
        )
        if i_end < i_start:
            raise ValueError(
                f"Cannot compute statistics on an empty interval."
                f" Requested : {ds.dates[i_start]} {ds.dates[i_end]}."
                f" Available: {ds.dates[0]=} {ds.dates[-1]=}"
            )

        stats_shape = (i_len, shape[1])

        mean = np.zeros(shape=stats_shape)
        stdev = np.zeros(shape=stats_shape)
        minimum = np.zeros(shape=stats_shape)
        maximum = np.zeros(shape=stats_shape)
        sums = np.zeros(shape=stats_shape)
        squares = np.zeros(shape=stats_shape)
        count = np.zeros(shape=stats_shape)

        for i, i_data in enumerate(range(i_start, i_end + 1)):
            chunk = data[i_data, ...]
            self.print(
                f"Computing statistics on {i_data+1}/{shape[0]} ({ds.dates[i_data]})"
            )
            for j in range(shape[1]):
                values = chunk[j, :]
                check_data_values(
                    values,
                    name=ds.variables[j],
                    log=[j, i, i_data, "statistics"],
                )
                minimum[i, j] = np.amin(values)
                maximum[i, j] = np.amax(values)
                sums[i, j] = np.sum(values)
                squares[i, j] = np.sum(values * values)
                count[i, j] = values.size
                mean[i, j] = sums[i, j] / count[i, j]
                stdev[i, j] = np.sqrt(
                    squares[i, j] / count[i, j] - mean[i, j] * mean[i, j]
                )

                check_stats(
                    minimum=minimum[i, j],
                    maximum=maximum[i, j],
                    mean=mean[i, j],
                    msg=f" for {j} {ds.variables[j]}",
                )

        assert (count == count[:][0]).all(), count

        assert len(minimum.shape) == 2
        assert minimum.shape[0] == i_len
        assert minimum.shape[1] == shape[1]

        _minimum = np.amin(minimum, axis=0)
        _maximum = np.amax(maximum, axis=0)
        _count = np.sum(count, axis=0)
        _sums = np.sum(sums, axis=0)
        _squares = np.sum(squares, axis=0)
        _mean = _sums / _count
        _stdev = np.sqrt(_squares / _count - _mean * _mean)

        stats = {
            "mean": _mean,
            "stdev": _stdev,
            "minimum": _minimum,
            "maximum": _maximum,
            "sums": _sums,
            "squares": _squares,
            "count": _count,
        }

        for v in stats.values():
            assert v.shape == stats["mean"].shape

        for i, name in enumerate(ds.variables):
            check_stats(**{k: v[i] for k, v in stats.items()}, msg=f"{i} {name}")

        return stats


class HDF5Loader:
    def __init__(self, *args, **kwargs):
        raise NotImplementedError()

    def append_array(self, *args, **kwargs):
        raise NotImplementedError("Appending do HDF5 not yet implemented")

    def create_array(
        self,
        dataset,
        shape,
        chunks,
        dtype,
        metadata,
        grid_points,
        nloops,
    ):
        import h5py

        if not isinstance(chunks, tuple):
            chunks = None

        print(
            f"Creating HDD5 file '{self.path}', with {dataset=}, {shape=}, {chunks=} and {dtype=}"
        )

        self.h5 = h5py.File(self.path, mode="w")
        array = self.h5.create_dataset(
            dataset,
            chunks=chunks,
            maxshape=shape,
            dtype=dtype,
            data=np.empty(
                shape
            )  # Can we avoid that? Looks like its needed for chuncking
            # data = h5py.Empty(dtype),
        )
        return array

    def close(self):
        self.h5.close()
        del self.h5

    def print_info(self):
        import h5py

        def h5_tree(h5, depth=0):
            for k, v in h5.items():
                if isinstance(v, h5py._hl.group.Group):
                    h5_tree(v, depth + 1)
                else:
                    print(" " * (depth * 3), k, v)
                    for p, q in v.attrs.items():
                        print(" " * (depth * 3 + 3), p, q)

        size = os.path.getsize(self.path)
        print(f"HDF5 file {self.path}: {size:,} ({bytes(size)})")
        with h5py.File(self.path, mode="r") as f:
            print("Content:")
            h5_tree(f, 1)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import fnmatch
import logging
import os
import shutil

from climetlab import load_source

from . import Reader
from . import reader as find_reader

LOG = logging.getLogger(__name__)


class GlobFilter:
    def __init__(self, pattern, top):
        self.pattern = pattern
        self.skip = len(top) + 1

    def __call__(self, path):
        match = fnmatch.fnmatch(path[self.skip :], self.pattern)
        LOG.debug("GlobFilter %s %s %s", path[self.skip :], self.pattern, match)
        return match


def make_file_filter(filter, top):
    if filter is None:
        return lambda _: True

    if callable(filter):
        return filter

    if isinstance(filter, str):
        return GlobFilter(filter, top)

    raise TypeError(f"Invalid filter {filter}")


class DirectoryReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)

        self._content = []

        filter = make_file_filter(self.filter, self.path)

        for root, _, files in os.walk(self.path):
            for file in files:
                full = os.path.join(root, file)
                LOG.debug("%s", full)
                if filter(full):
                    self._content.append(full)

    def mutate(self):
        if len(self._content) == 1:
            return find_reader(self.source, self._content[0])
        return self

    def mutate_source(self):
        if os.path.exists(os.path.join(self.path, ".zattrs")):
            return load_source("zarr", self.path)

        return load_source(
            "multi",
            [
                load_source(
                    "file",
                    path=path,
                    filter=self.filter,
                    merger=self.merger,
                )
                for path in sorted(self._content)
            ],
            filter=self.filter,
            merger=self.merger,
        )

    def save(self, path):
        shutil.copytree(self.path, path)

    def write(self, f):
        raise NotImplementedError()


def reader(source, path, magic=None, deeper_check=False):
    if magic is None or os.path.isdir(path):
        return DirectoryReader(source, path)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import re
import weakref
from importlib import import_module

from climetlab.core import Base
from climetlab.core.caching import cache_file
from climetlab.core.plugins import find_plugin
from climetlab.core.plugins import register as register_plugin
from climetlab.core.settings import SETTINGS
from climetlab.utils.html import table


class Source(Base):
    """
    Doc
    """

    name = None
    home_page = "-"
    licence = "-"
    documentation = "-"
    citation = "-"

    _dataset = None
    _parent = None

    def __init__(self, **kwargs):
        self._kwargs = kwargs

    def settings(self, name):
        return SETTINGS.get(name)

    def mutate(self):
        # Give a chance to `multi` to change source
        return self

    def ignore(self):
        # Used by multi-source
        return False

    def __add__(self, other):
        from climetlab.sources import load_source

        return load_source("multi", self, other)

    def cache_file(self, create, args, **kwargs):
        owner = self.name
        if self.dataset:
            owner = self.dataset.name
        if owner is None:
            owner = re.sub(r"(?!^)([A-Z]+)", r"-\1", self.__class__.__name__).lower()

        resource = None
        for connection in self.connect_to_mirrors():
            resource = connection.get_file(create, args)
        if resource:
            return resource

        return cache_file(owner, create, args, **kwargs)

    @property
    def dataset(self):
        if self._dataset is None:
            return None
        return self._dataset()

    @dataset.setter
    def dataset(self, dataset):
        self._set_dataset(weakref.ref(dataset))

    def _set_dataset(self, dataset):
        self._dataset = dataset

    @property
    def parent(self):
        if self._parent is None:
            return None
        return self._parent()

    @parent.setter
    def parent(self, parent):
        self._set_parent(weakref.ref(parent))

    def _set_parent(self, parent):
        self._parent = parent

    def _repr_html_(self):
        return table(self)

    def graph(self, depth=0):
        print(" " * depth, self)

    # Mirroring
    def connect_to_mirrors(self):
        result = []
        from climetlab.mirrors import get_active_mirrors

        for mirror in get_active_mirrors():
            c = self.connect_to_mirror(mirror)
            if c is None:
                continue
            result.append(c)
        return result

    def connect_to_mirror(self, mirror):
        return None


class SourceLoader:
    kind = "source"

    def load_module(self, module):
        return import_module(module, package=__name__).source

    def load_entry(self, entry):
        entry = entry.load()
        if callable(entry):
            return entry
        return entry.source

    def load_remote(self, name):
        return None


class SourceMaker:
    def __call__(self, name, *args, **kwargs):
        loader = SourceLoader()

        if ("." in name or "/" in name) and os.path.exists(name):
            raise ValueError(
                (
                    f'"{name}" is not a valid source name, but the file exists. '
                    'Did you mean load_source("file", "{name}") ?'
                )
            )

        if name.startswith("http://") or name.startswith("https://"):
            raise ValueError(
                f'"{name}" is not a valid source name. Did you mean load_source("url", "{name}") ?'
            )

        klass = find_plugin(os.path.dirname(__file__), name, loader)

        if os.environ.get("CLIMETLAB_TESTING_ENABLE_MOCKUP_SOURCE", False):
            from climetlab.mockup import SourceMockup

            klass = SourceMockup

        source = klass(*args, **kwargs)

        if getattr(source, "name", None) is None:
            source.name = name

        return source

    def __getattr__(self, name):
        return self(name.replace("_", "-"))


get_source = SourceMaker()


def load_source(name: str, *args, lazily=False, **kwargs) -> Source:
    if lazily:
        return load_source_lazily(name, *args, **kwargs)

    prev = None
    src = get_source(name, *args, **kwargs)
    while src is not prev:
        prev = src
        src = src.mutate()
    return src


def load_source_lazily(name, *args, **kwargs):
    from climetlab.utils.lazy import LazySource

    return LazySource(name, *args, **kwargs)


def register(name, proc):
    register_plugin("source", name, proc)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from importlib import import_module

import yaml

import climetlab
from climetlab.core import Base
from climetlab.core.metadata import annotate
from climetlab.core.plugins import find_plugin
from climetlab.core.plugins import register as register_plugin
from climetlab.core.settings import SETTINGS
from climetlab.utils import download_and_cache
from climetlab.utils.html import table

LOG = logging.getLogger(__name__)


class Dataset(Base):
    """Mother class to create a dataset object."""

    name = None
    """ str : To be overrided by the class inherithing from Dataset.

    The name of the dataset.
    """

    home_page = "-"
    """ str : To be overrided by the class inherithing from Dataset.

    Contains a link to the home page related to the dataset.
    """

    licence = "-"
    """ str : To be overrided by the class inherithing from Dataset.

    Contains a link to the licence of the dataset.
    """

    documentation = "-"
    """ str : To be overrided by the class inherithing from Dataset.

    Contains a link to the documentation related to the dataset.
    """

    citation = "-"
    """ str : To be overrided by the class inherithing from Dataset.

    Contains the citation related to the dataset.
    """

    terms_of_use = None
    """ str : To be overrided by the class inherithing from Dataset.

    Contains the Terms of Use of the dataset.
    It will be shown to the user when they download the dataset for the first time.

    Uses :py:const:`TERMS_OF_USE_SHOWN`
    """

    _source = None

    def __init__(self, *args, **kwargs):
        """Do nothing. To be overridden by the inherithing class."""
        pass

    def mutate(self):
        """Give a chance to a subclass to change itself to another class after creation time.

        Returns
        -------
        self
        """
        return self

    def __add__(self, other):
        from climetlab.sources import load_source

        return load_source("multi", self.source, other.source)

    @property
    def source(self):
        """Most methods are delegated to the ``source`` of the dataset.
        Returns
        -------
        :py:class:`Source`
        """
        return self._source

    @source.setter
    def source(self, source):
        self._source = source
        source.dataset = self

    def __len__(self):
        return len(self.source)

    def __getitem__(self, n):
        return self.source[n]

    def sel(self, *args, **kwargs):
        return self.source.sel(*args, **kwargs)

    def order_by(self, *args, **kwargs):
        return self.source.order_by(*args, **kwargs)

    def statistics(self, *args, **kwargs):
        return self.source.statistics(*args, **kwargs)

    @property
    def coords(self):
        return self.source.coords

    @property
    def all_coords(self):
        return self.source.all_coords

    def to_numpy(self, **kwargs):
        import numpy as np

        return np.array([s.to_numpy(**kwargs) for s in self.source])

    def to_xarray(self, **kwargs):
        return self.source.to_xarray(**kwargs)

    def to_tfdataset(self, **kwargs):
        return self.source.to_tfdataset(**kwargs)

    def to_pandas(self, **kwargs):
        return self.source.to_pandas(**kwargs)

    def to_metview(self, **kwargs):
        return self.source.to_metview(**kwargs)

    def _repr_html_(self):
        return table(self)

    def annotate(self, data, **kargs):
        return annotate(data, self, **kargs)

    def ignore(self):
        return self.source.ignore()

    def unique_values(self, *args, **kwargs):
        return self.source.unique_values(*args, **kwargs)


def _module_callback(plugin):
    return import_module(plugin, package=__name__).dataset


def camel(name):
    return "".join(x[0].upper() + x[1:].lower() for x in name.split("-"))


class YamlDefinedDataset(Dataset):
    """Dataset class to defined datasets from a YAML file.

    When inheriting, self._src and self._args must be defined before calling __init__.
    This is performed in :py:class:`DatasetLoader`.

    """

    def __init__(self):
        self.source = climetlab.load_source(self._src, **self._args)

    def __call__(self):
        return self


def _dataset_from_dict(name, dataset, path=None):
    attributes = dataset.get("metadata", {})
    attributes.update(
        dict(_path=path, _src=dataset["source"], _args=dataset.get("args", {}))
    )
    return type(camel(name), (YamlDefinedDataset,), attributes)


def dataset_from_dict(path, *args, **kwargs):
    return _dataset_from_dict(path)(*args, **kwargs).mutate()


def _dataset_from_yaml(path):
    name, _ = os.path.splitext(os.path.basename(path))
    with open(path) as f:
        dataset = yaml.load(f.read(), Loader=yaml.SafeLoader)["dataset"]
        return _dataset_from_dict(name, dataset, path)


def dataset_from_yaml(path, *args, **kwargs):
    return _dataset_from_yaml(path)(*args, **kwargs).mutate()


class DatasetLoader:
    kind = "dataset"

    def settings(self, name):
        return SETTINGS.get(name)

    def load_yaml(self, path):
        return _dataset_from_yaml(path)

    def load_module(self, module):
        return import_module(module, package=__name__).dataset

    def load_entry(self, entry):
        entry = entry.load()
        if callable(entry):
            return entry
        return entry.dataset

    def load_remote(self, name):
        catalogs = self.settings("datasets-catalogs-urls")
        for catalog in catalogs:
            url = f"{catalog}/{name}.yaml"
            path = download_and_cache(
                url,
                update_if_out_of_date=True,
                return_none_on_404=True,
            )

            if path:
                LOG.debug("Found dataset at %s", url)
                return self.load_yaml(path)

        return None


class DatasetMaker:
    def lookup(self, name):
        loader = DatasetLoader()
        klass = find_plugin(os.path.dirname(__file__), name, loader)

        return klass

    def __call__(self, name, *args, **kwargs):
        return self.lookup(name, *args, **kwargs)

    def __getattr__(self, name):
        if name.startswith("_"):
            raise AttributeError(name)
        return self(name.replace("_", "-"))


get_dataset = DatasetMaker()


TERMS_OF_USE_SHOWN = set()


def load_dataset(name: str, *args, **kwargs) -> Dataset:
    """Loads a dataset.

    Parameters
    ----------
    name : str
        Name of the dataset to be loaded.

    Returns
    -------
    Dataset
        The loaded dataset.
    """
    klass = get_dataset.lookup(name)

    if name not in TERMS_OF_USE_SHOWN:
        if hasattr(klass, "terms_of_use") and klass.terms_of_use is not None:
            print(klass.terms_of_use)
        TERMS_OF_USE_SHOWN.add(name)

    ds = klass(*args, **kwargs).mutate()
    if getattr(ds, "name", None) is None:
        ds.name = name
    return ds


def register(name, proc):
    register_plugin("dataset", name, proc)



#!/usr/bin/env python
#
# (C) Copyright 2021- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.
#

import io
import itertools
import json
import os
import pickle

import yaml

from climetlab.utils import load_json_or_yaml
from climetlab.utils.factorise import Tree, factorise

from .humanize import dict_to_human, list_to_human


def _tidy_dict(query):
    result = dict()
    for k, v in query.items():
        if v is None:
            continue
        result[k] = v
    return result


def load_str(avail):
    try:
        return json.loads(avail)
    except json.decoder.JSONDecodeError:
        return yaml.safe_load(avail)


def load_json(avail):
    with open(avail) as f:
        return json.loads(f.read())


def load_yaml(avail):
    with open(avail) as f:
        return yaml.load(f, Loader=yaml.SafeLoader)


class Availability:
    def __init__(self, avail, intervals=None, parser=None):
        CONFIG_LOADERS = {
            "json": load_json,
            "yaml": load_yaml,
            # "marslist": Availability.from_mars_list,
            "pickle": Availability.from_pickle,
        }
        if not isinstance(avail, Tree):
            if isinstance(avail, str):
                config_loader = load_json
                if len(avail) > 5:
                    ext = avail[-10:].split(".")[-1]
                    config_loader = CONFIG_LOADERS.get(ext, load_str)
                    avail = config_loader(avail)

            if parser is not None:
                avail = [parser(item) for item in avail]

            if isinstance(avail, Availability):
                avail = avail._tree

            if not isinstance(avail, Tree):
                avail = factorise(avail, intervals=intervals)
        self._tree = avail

    def as_mars_list(self, *args, **kwargs):
        return self._tree.as_mars_list(*args, **kwargs)

    def as_mars(self, *args, **kwargs):
        return self._tree.as_mars(*args, **kwargs)

    def to_pickle(self, filename):
        with open(filename, "wb") as f:
            pickle.dump(self._tree, f)

    @classmethod
    def from_pickle(cls, filename):
        with open(filename, "rb") as f:
            tree = pickle.load(f)
            return cls(tree)

    def to_yaml(self):
        return yaml.dump(self._tree)

    @classmethod
    def from_yaml(cls, filename):
        dic = load_json_or_yaml(filename)
        return cls(dic)

    @classmethod
    def from_mars_list(cls, tree, intervals=None):
        if os.path.exists(tree):
            input = open(tree)
        else:
            input = io.StringIO(tree)

        def as_dict(s):
            r = {}
            for a in s.split(","):
                p, v = a.split("=")
                r[p] = v.split("/")
            return r

        requests = []
        stack = []
        last = -1
        for line in input:
            line = line.rstrip()
            cnt = 0
            while len(line) > 0 and line[0] == " ":
                line = line[1:]
                cnt += 1
            if cnt <= last:
                requests.append(as_dict(",".join(stack)))
            while len(stack) > cnt:
                stack.pop()
            stack.append(line)
            last = cnt

        if stack:
            requests.append(as_dict(",".join(stack)))

        return cls(requests, intervals)

    def _repr_html_(self):
        return "<hr><pre>{}</pre><hr>".format(self.tree())

    def __str__(self):
        return self.tree()

    def __repr__(self):
        return str(self.tree())

    def select(self, *args, **kwargs):
        return Availability(self._tree.select(*args, **kwargs))

    def missing(self, *args, **kwargs):
        # assert kwargs.keys() == self.unique_values().keys(), "kwargs must contain all dimensions"
        return Availability(self._tree.missing(*args, **kwargs))

    def check(self, _kwargs=None, **kwargs):
        assert _kwargs is None or not kwargs

        if _kwargs is not None and not kwargs:
            assert isinstance(_kwargs, dict)
            kwargs = _kwargs

        if self.count(kwargs):
            return

        reasons = []

        #         u = self.unique_values()
        #         for k, v in kwargs.items():
        #             if v is None:
        #                 continue
        #             if k not in u:
        #                 reasons.append(f"Unknown key {k}")
        #                 continue
        #             if v not in u[k]:
        #                 reasons.append(f"Invalid value for {k}: {v} must be in {u[k]}")
        #                 continue

        query = kwargs

        if not reasons:

            def iterate_request(r):
                yield from (
                    dict(zip(r.keys(), x)) for x in itertools.product(*r.values())
                )

            def build(x):
                # if isinstance(x, (list, tuple)):
                #    return [None] + list(x)
                return [None, x]

            r = {k: build(v) for k, v in query.items()}

            lst = []
            for i in iterate_request(r):
                if self.count(i) == 0:
                    i = _tidy_dict(i)
                    lst.append((abs(len(i) - 2), i))

            lst = sorted(lst, key=lambda x: x[0])

            if len(lst) > 0:
                for x in lst:
                    if x[0] == lst[0][0]:
                        ii = dict_to_human(x[1])
                        reasons.append(f"invalid combination ({ii})")

        raise ValueError(f"{list_to_human(reasons)}.")

    def __len__(self):
        return self.count()

    def __getattr__(self, name):
        return getattr(self._tree, name)


if __name__ == "__main__":
    for n in Availability.from_mars_list("mars.list.tree").iterate():
        print(n)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import hashlib
import json
import logging
import os
import sqlite3
import time
from threading import local

import numpy as np

import climetlab as cml
from climetlab.indexing.database.json import json_serialiser
from climetlab.loaders import build_remapping
from climetlab.utils import tqdm
from climetlab.utils.parts import Part

from . import (
    FILEPARTS_KEY_NAMES,
    MORE_KEY_NAMES,
    MORE_KEY_NAMES_WITH_UNDERSCORE,
    STATISTICS_KEY_NAMES,
    Database,
    FloatDBKey,
    IntDBKey,
    StrDBKey,
)

LOG = logging.getLogger(__name__)


def dump_sql(statement):
    statement = statement.replace(",", ", ")
    statement = statement.replace(";", " ;")
    statement = statement.replace("(", " (")
    lst = statement.split()
    lst = [
        "userorder_entries_...\n" if x.startswith("userorder_entries") else x
        for x in lst
    ]
    print(" ".join(lst))


def execute(connection, statement, *arg, **kwargs):
    if LOG.level == logging.DEBUG:
        assert False
        dump_sql(statement)

    delay = 1
    while delay < 30 * 60:  # max delay 30 min
        try:
            return connection.execute(statement, *arg, **kwargs)
        except sqlite3.OperationalError as e:
            if not str(e).endswith("database is locked"):
                raise e
            time.sleep(delay)
            dump_sql(statement)
            print(f"{e}. Retrying in {delay} seconds.")
            delay = delay * 1.5
    raise e  # noqa: F821


def entryname_to_dbname(n):
    n = dict(
        levellist="levelist",
        level="levelist",
        leveltype="levtype",
        variable="param",
        parameter="param",
        realization="number",
        realisation="number",
        klass="class",
    ).get(n, n)

    def add_mars(x):
        return "i_" + x

    def remove_first_underscore(x):
        assert x[0] == "_", x
        return x[1:]

    if n in FILEPARTS_KEY_NAMES:
        return remove_first_underscore(n)
    if n in MORE_KEY_NAMES_WITH_UNDERSCORE:
        return remove_first_underscore(n)
    if n in STATISTICS_KEY_NAMES:
        return n
    # if n in MORE_KEY_NAMES:
    #    return n
    return add_mars(n)


def dbname_to_entryname(n):
    if n.startswith("i_"):
        return n[2:]
    if "_" + n in FILEPARTS_KEY_NAMES:
        return "_" + n
    if "_" + n in MORE_KEY_NAMES_WITH_UNDERSCORE:
        return "_" + n
    if n in MORE_KEY_NAMES:
        return n
    return n


class EntriesLoader:
    table_name = "entries"

    def __init__(self, connection):
        self.connection = connection
        self.patch()
        self.keys = self.read_from_table()
        self.path_table = PathTable(connection)

    def patch(self):
        try:
            execute(
                self.connection,
                "ALTER TABLE entries RENAME COLUMN i_datetime TO i_valid_datetime;",
            )
        except sqlite3.OperationalError:
            pass

        try:
            execute(
                self.connection, "ALTER TABLE entries DROP COLUMN i_param_levelist;"
            )
        except sqlite3.OperationalError:
            pass

    @classmethod
    def guess_key_type(self, x, name=None):
        try:
            return {
                "TEXT": StrDBKey,
                "FLOAT": FloatDBKey,
                "INTEGER": IntDBKey,
                str: StrDBKey,
                float: FloatDBKey,
                np.float64: FloatDBKey,
                np.float32: FloatDBKey,
                int: IntDBKey,
                datetime.datetime: StrDBKey,
            }[x]
        except KeyError:
            raise ValueError(f"Unknown type '{x}' for key '{name}'.")

    def read_from_table(self):
        keys = {}
        cursor = execute(self.connection, f"PRAGMA table_info({self.table_name})")
        for x in cursor.fetchall():
            name, typ = x[1], x[2]
            klass = self.guess_key_type(typ, name)
            keys[name] = klass(name)

        if not keys:
            LOG.debug(f"Table {self.table_name} does not exist.")
        return keys

    def create_table_from_entry_if_needed(self, entry):
        keys = {}
        for k, v in entry.items():
            dbkey = self._build_dbkey(k, v)
            keys[dbkey.name] = dbkey

        assert keys, f"Cannot build from entry '{entry}'"
        LOG.debug(f"Created table {self} from entry {entry}.")
        columns_defs = ",".join([f"{v.name} {v.sql_type}" for k, v in keys.items()])
        statement = f"CREATE TABLE IF NOT EXISTS {self.table_name} ({columns_defs});"
        execute(self.connection, statement)
        return self.read_from_table()

    def _add_column(self, k, v):
        dbkey = self._build_dbkey(k, v)
        statement = (
            f"ALTER TABLE {self.table_name} ADD COLUMN {dbkey.name} {dbkey.sql_type};"
        )
        LOG.debug("%s", statement)
        try:
            execute(self.connection, statement)
        except sqlite3.OperationalError:
            LOG.debug("Add column failed, this is expected because of concurency")
        self.keys[dbkey.name] = dbkey
        return self.keys

    def _build_dbkey(self, k, v):
        typ = type(v)
        klass = self.guess_key_type(typ, name=k)
        name = entryname_to_dbname(k)
        return klass(name)

    def load_iterator(self, iterator):
        paths_or_urls = set()

        count = 0
        for entry in iterator:
            if count == 0:
                self.keys = self.create_table_from_entry_if_needed(entry)

            for k, v in entry.items():
                dbname = entryname_to_dbname(k)
                if dbname not in self.keys:
                    LOG.debug(f"Inserting column in databse {k}, {dbname}")
                    self.keys = self._add_column(k, v)

            if "_path" in entry:
                paths_or_urls.add(entry["_path"])
            if "_url" in entry:
                paths_or_urls.add(entry["_url"])

            column_names = [k for k, v in self.keys.items()]
            statement = (
                f"INSERT INTO {self.table_name} ("
                + ",".join(column_names)
                + ") VALUES("
                + ",".join(["?"] * len(column_names))
                + ");"
            )
            values = [entry.get(dbname_to_entryname(k)) for k, v in self.keys.items()]
            execute(self.connection, statement, tuple(values))
            count += 1

        date = datetime.datetime.now().isoformat()
        for path in paths_or_urls:
            self.path_table.insert(path, date)

        return count

    def build_sql_indexes(self):
        indexed_columns = [k for k, v in self.keys.items() if k.startswith("i_")]
        indexed_columns += ["path"]

        pbar = tqdm(indexed_columns, desc="Building indexes")
        for n in pbar:
            pbar.set_description(f"Building index for {n}")
            execute(
                self.connection,
                f"CREATE INDEX IF NOT EXISTS {n}_index ON {self.table_name} ({n});",
            )

    def __str__(self):
        content = ",".join([k for k, v in self.keys.items()])
        return f"{self.__class__.__name__}({self.table_name},{content}"


class Connection(local):
    # Inheriting from threading.local allows one connection for each thread
    # __init__ is "called each time the local object is used in a separate thread".
    # https://github.com/python/cpython/blob/0346eddbe933b5f1f56151bdebf5bd49392bc275/Lib/_threading_local.py#L65
    def __init__(self, db_path):
        self._conn = sqlite3.connect(db_path)


class SqlFilter:
    def __init__(self, kwargs=None, remapping=None):
        self.kwargs = kwargs or {}
        self.remapping = build_remapping(remapping)

    def h(self, *args, **kwargs):
        m = hashlib.md5()
        m.update(str(args).encode("utf-8"))
        m.update(str(kwargs).encode("utf-8"))
        m.update(str(self.remapping.as_dict()).encode("utf-8"))
        m.update(str(self.__class__.__name__).encode("utf-8"))
        m.update(
            json.dumps(self.kwargs, sort_keys=True, default=json_serialiser).encode(
                "utf-8"
            )
        )
        return m.hexdigest()

    def __str__(self):
        return f"{self.__class__.__name__}({self.kwargs})"

    @property
    def is_empty(self):
        return not self.kwargs

    def create_new_view(self, db, view):
        new_view = "entries_" + self.h(parent_view=view)
        assert new_view != view
        view_statement = self.create_view_statement(
            db, old_view=view, new_view=new_view
        )
        if not view_statement:
            # nothing to do
            return view

        LOG.debug("%s", view_statement)
        for i in execute(db.connection, view_statement):
            LOG.error(str(i))
        return new_view


class SqlSelection(SqlFilter):
    def create_view_statement(self, db, old_view, new_view):
        conditions = []
        for k, v in self.kwargs.items():
            if v is None or v is cml.ALL:
                continue

            name = entryname_to_dbname(k)
            dbkey = db.dbkeys[name]

            if not isinstance(v, (list, tuple)):
                v = [v]

            v = [dbkey.cast(x) for x in v]
            v = ["'" + str(x) + "'" for x in v]

            conditions.append(f"{name} IN ({', '.join(v)})")

        if not conditions:
            return None

        assert new_view != old_view
        return (
            f"CREATE TEMP VIEW IF NOT EXISTS {new_view} AS SELECT * "
            + f"FROM {old_view} WHERE "
            + " AND ".join(conditions)
            + ";"
        )


class SqlRemapping(SqlFilter):
    def create_view_statement(self, db, old_view, new_view):
        class SqlCustomJoiner:
            def format_name(self, x):
                name = entryname_to_dbname(x)
                return f"COALESCE({name},'')"

            def format_string(self, name):
                if not name:
                    return name
                return "'" + str(name) + "'"

            def join(self, lst):
                lst = [_ for _ in lst if len(_)]
                assert len(lst) > 0, lst
                return " || ".join(lst)

        sql_concatenations = []
        for k, v in self.remapping.remapping.items():
            joiner = SqlCustomJoiner()
            alias = entryname_to_dbname(k)
            expr = self.remapping.substitute(k, joiner)
            sql_concatenations.append(f"TRIM({expr},'_') AS {alias}")

        select = ", ".join(sql_concatenations)

        if not sql_concatenations:
            return None

        assert new_view != old_view
        return (
            f"CREATE TEMP VIEW IF NOT EXISTS {new_view} AS SELECT *, {select} "
            f"FROM {old_view};"
        )


class SqlOrder(SqlFilter):
    def merge(self, *others):
        orders = reversed([self, *others])

        kwargs = {}
        for o in orders:
            for k, v in o.kwargs.items():
                if k not in kwargs:
                    kwargs[k] = v

        return SqlOrder(kwargs)

    def create_view_statement(self, db, old_view, new_view):
        func_name = "userorder_" + new_view

        order_bys = []
        dict_of_dicts = {}

        for k, v in self.kwargs.items():
            name = entryname_to_dbname(k)

            dbkey = db.dbkeys.get(name, StrDBKey(name))

            if v == "ascending" or v is None:
                order_bys.append(name + " ASC")
                continue
            if v == "descending":
                order_bys.append(name + " DESC")
                continue
            if isinstance(v, (list, tuple)):
                v = [dbkey.cast(x) for x in v]
                dict_of_dicts[name] = dict(zip(v, range(len(v))))
                order_bys.append(func_name + "(" + "'" + name + "'" + ", " + name + ")")
                continue

            raise ValueError(f"{k},{v}, {type(v)}")

        if not order_bys:
            return None

        if dict_of_dicts:

            def order_func(k, v):
                return dict_of_dicts[k][v]

            if LOG.level == logging.DEBUG:
                sqlite3.enable_callback_tracebacks(True)
            db.connection.create_function(func_name, 2, order_func, deterministic=True)

        assert new_view != old_view
        return (
            f"CREATE TEMP VIEW IF NOT EXISTS {new_view} AS SELECT * "
            + f"FROM {old_view} ORDER BY "
            + ",".join(order_bys)
            + ";"
        )


class VersionedDatabaseMixin:
    VERSION = 6

    @property
    def _version(self):
        cursor = execute(self.connection, "PRAGMA user_version;")
        for res in cursor:
            version = res[0]
            return version if version else None
        assert False

    def _set_version(self):
        if self._version is None:
            execute(self.connection, f"PRAGMA user_version = {self.VERSION};")
            return
        self._check_version()

    def _check_version(self):
        version = self._version
        if version is None or version == self.VERSION:
            return
        raise Exception(
            (
                "Version mismatch: current version for database index"
                " is {self.VERSION} and the database already has version"
                f" {version}"
            )
        )


class PathTable:
    table_name = "paths"

    def __init__(self, connection):
        self.connection = connection
        self.ensure_table()

    def ensure_table(self):
        statement = f"CREATE TABLE IF NOT EXISTS {self.table_name} (key TEXT PRIMARY KEY, date TEXT);"
        for i in execute(self.connection, statement):
            LOG.error(str(i))  # Output of .execute should be empty

    def insert(self, key, date):
        statement = f"""INSERT INTO {self.table_name} (key, date) VALUES(?,?);"""
        LOG.debug("%s", statement)
        execute(self.connection, statement, (key, date))

    def get_date(self, key):
        statement = f"""SELECT * FROM {self.table_name} WHERE key="{key}";"""
        LOG.debug("%s", statement)
        for key, date in execute(self.connection, statement):
            return date
        return None


class SqlDatabase(Database, VersionedDatabaseMixin):
    EXTENSION = ".db"

    def __init__(
        self,
        db_path,
        filters=None,
    ):
        self._cache_column_names = {}

        self.db_path = db_path
        self._filters = filters or []
        self._view = None
        self._connection = None

        self.dbkeys = EntriesLoader(self.connection).keys

    def __str__(self):
        return f"{self.__class__.__name__}({self.db_path},filters=[{','.join([str(_) for _ in self._filters])}])"

    def build_indexes(self):
        EntriesLoader(self.connection).build_sql_indexes()

    @property
    def view(self):
        if self._view is None:
            self._view = "entries"
            for f in self._filters:
                self._view = f.create_new_view(self, self._view)
            LOG.debug("DB %s %s", self.db_path, self._view)
        return self._view

    @property
    def connection(self):
        if self._connection is None:
            self._connection = Connection(self.db_path)
        return self._connection._conn

    def unique_values(self, *coords, remapping=None, progress_bar=True):
        """
        Given a list of metadata attributes, such as date, param, levels,
        returns the list of unique values for each attributes
        """

        remapping = build_remapping(remapping)
        with self.connection as con:
            view = self.view
            # orders = [f for f in self._filters if isinstance(f, SqlOrder)]
            # if orders:
            #     order = orders[0].merge(*orders)
            #     view = order.create_new_view(self, view)

            results = {}
            for c in coords:
                column = entryname_to_dbname(c)
                values = [
                    v[0] for v in execute(con, f"SELECT DISTINCT {column} FROM {view};")
                ]
                LOG.debug("Reordered values for {column}", column, values)
                results[column] = values

        return results

    def filter(self, filter: SqlFilter):
        return self.__class__(
            self.db_path,
            filters=self._filters + [filter],
        )

    def already_loaded(self, path_or_url, owner):
        with self.connection as connection:
            date = PathTable(connection).get_date(path_or_url)
            return date is not None

    def load_iterator(self, iterator):
        with self.connection as connection:
            loader = EntriesLoader(connection)
            count = loader.load_iterator(iterator)
            self.dbkeys = loader.keys

            assert count >= 1, "No entry found."
            LOG.info("Added %d entries", count)

        return count

    def lookup_parts(self, limit=None, offset=None, resolve_paths=True):
        """
        Look into the database and provide entries as Parts.
        limit: Returns only "limit" entries (used for paging).
        offset: Skip the first "offset" entries (used for paging).
        """

        _names = ["path", "offset", "length"]
        parts = []
        for path, offset, length in self._execute_select(_names, limit, offset):
            parts.append(Part(path, offset, length))
        if resolve_paths:
            parts = Part.resolve(parts, os.path.dirname(self.db_path))
        return parts

    def lookup_dicts(self, limit=None, offset=None, remove_none=True, with_parts=None):
        """
        Return entries dicts as they where inserted into the database.
        limit: Returns only "limit" entries (used for paging).
        offset: Skip the first "offset" entries (used for paging).
        """

        column_names = [k for k, v in self.dbkeys.items()]

        if with_parts is False:
            column_names = [k for k in column_names if k not in FILEPARTS_KEY_NAMES]

        entrynames = [dbname_to_entryname(k) for k in column_names]
        for tupl in self._execute_select(column_names, limit, offset):
            dic = {k: v for k, v in zip(entrynames, tupl)}

            if remove_none:
                dic = {k: v for k, v in dic.items() if v is not None}
            yield dic

    def _execute_select(self, column_names, limit=None, offset=None):
        names_str = ",".join([x for x in column_names]) if column_names else "*"
        limit_str = f" LIMIT {limit}" if limit is not None else ""
        offset_str = f" OFFSET {offset}" if offset is not None else ""

        statement = f"SELECT {names_str} FROM {self.view} {limit_str} {offset_str};"
        LOG.debug("%s", statement)

        for tupl in execute(self.connection, statement):
            yield tupl

    def count(self):
        statement = f"SELECT COUNT(*) FROM {self.view};"
        for result in execute(self.connection, statement):
            return result[0]
        assert False, statement  # Fail if result is empty.

    def duplicate_db(self, filename, **kwargs):
        new_db = SqlDatabase(db_path=filename)
        iterator = self.lookup_dicts()
        new_db.load(iterator)
        return new_db



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging

from climetlab.arguments.climetlab_types import infer_type
from climetlab.arguments.transformers import (
    AliasTransformer,
    FormatTransformer,
    TypeTransformer,
)

LOG = logging.getLogger(__name__)


def check_included(values1, values2):
    if isinstance(values1, (list, tuple)) and isinstance(values2, (list, tuple)):
        for x in values1:
            if x not in values2:
                raise ValueError(f"'{x}' is not in {values2}")


class Argument:
    def __init__(
        self,
        name,
    ):
        assert name != "format"
        self.name = name
        self.availability = None
        self._type = None
        self._normalize = {}
        self.format = None
        self.aliases = None

    @property
    def normalize(self):
        return self._normalize

    @normalize.setter
    def normalize(self, value):
        self.format = value.pop("format", None)
        self.aliases = value.pop("aliases", None)
        self._normalize = value

    @property
    def cmltype(self):
        if self._type is None:
            if self.availability:
                if self._normalize.get("values") is None:
                    self._normalize["values"] = self.availability
            self._type = infer_type(**self.normalize)
        return self._type

    def add_alias_transformers(self, pipeline):
        if self.aliases:
            pipeline.append(AliasTransformer(self, self.cmltype, self.aliases))

    def add_type_transformers(self, pipeline):
        pipeline.append(TypeTransformer(self, self.cmltype))

    def add_format_transformers(self, pipeline):
        if self.format is not None:
            pipeline.append(FormatTransformer(self, self.format, self.cmltype))

    def __repr__(self) -> str:
        return f"Argument({self.name})"



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


from climetlab.readers.grib.index.sql import FieldsetInFilesWithSqlIndex
from climetlab.sources.indexed import IndexedSource
from climetlab.sources.indexed_urls import add_path, get_index_url


class IndexedUrl(IndexedSource):
    def __init__(
        self,
        url,
        substitute_extension=False,
        index_extension=".index",
        **kwargs,
    ):
        index = FieldsetInFilesWithSqlIndex.from_url(
            get_index_url(url, substitute_extension, index_extension),
            patch_entry=add_path(url),
        )
        super().__init__(index, **kwargs)


source = IndexedUrl



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import xarray as xr
import zarr

from . import Source

LOG = logging.getLogger(__name__)


class Cache:
    def __init__(self, store):
        self._store = store

    def __contains__(self, key):
        # print("Cache.__contains__", key)
        return key in self._store

    def __getitem__(self, key):
        # print("Cache.__getitem__", key)
        data = self._store[key]
        # print(len(data))
        return data

    def keys(self):
        return self._store.keys()


class ZarrS3(Source):
    def __init__(self, urls, **kwargs):
        super().__init__(**kwargs)

        if not isinstance(urls, list):
            urls = [urls]

        def url_to_store(url):
            import s3fs

            bits = url.split("/")

            url = "/".join(bits[:3])
            root = "/".join(bits[3:])

            fs = s3fs.S3FileSystem(anon=True, client_kwargs={"endpoint_url": url})

            store = s3fs.S3Map(
                root=root,
                s3=fs,
                check=False,
            )

            store = Cache(store)

            store = zarr.storage.KVStore(store)

            return store

        # adding a new dimension take a lot of memory
        # dslist = [xr.open_dataset(url_to_store(url), engine="zarr") for url in urls]
        # self._ds = xr.concat(dslist, dim = 'head_time')
        # return self._ds

        # options = self.read_zarr_options()

        # concat_dim = options.get("concat_dim", "forecast_time")
        concat_dim = "forecast_time"  # TODO: fix me

        stores = [url_to_store(url) for url in urls]

        dslist = []
        for store, url in zip(stores, urls):
            try:
                dslist.append(xr.open_dataset(store, engine="zarr", chunks="auto"))
            except zarr.errors.GroupNotFoundError as e:
                print(f"ERROR : Cannot find data at url = {url}")
                raise (e)

        assert len(dslist) > 0

        if len(dslist) == 1:
            self._ds = dslist[0]
        else:
            dsdict = {}
            for ds in dslist:
                for value in ds[concat_dim].values:
                    dsdict[value] = ds.sel(**{concat_dim: value})
                values_sorted = sorted(dsdict.keys())
                dslist = [dsdict[d] for d in values_sorted]

            self._ds = xr.concat(dslist, dim=concat_dim)
            # self._ds = xr.open_mfdataset(stores, engine="zarr", combine='nested')

    def to_xarray(self):
        # self._ds = self.post_xarray_open_dataset_hook(self._ds)
        return self._ds


source = ZarrS3



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.sources import Source


class TestingMockup:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class TestingXarrayAttrs(dict):
    pass


class TestingXarrayDims(list):
    pass


class TestingDatasetAsXarray(TestingMockup):
    def __init__(self, *args, **kwargs):
        super(TestingDatasetAsXarray, self).__init__(*args, **kwargs)
        self.attrs = TestingXarrayAttrs()
        self.dims = TestingXarrayDims()

    # TODO: make this generic
    def min(self, *args, **kwargs):
        print(f"xr.min({args}, {kwargs})")
        return 42.0

    def max(self, *args, **kwargs):
        print(f"xr.min({args}, {kwargs})")
        return 42.0

    def map(self, *args, **kwargs):
        print("xr.map(...)")
        # print(f'xr.map({args}, {kwargs})')
        return self

    def sortby(self, *args, **kwargs):
        print(f"xr.sortby({args}, {kwargs})")
        return self

    def __getitem__(self, key):
        print(f"xr.__getitem__({key})")
        return self

    def __setitem__(self, key, value):
        print(f"xr.__setitem__({key})=...")
        # print(f'xr.__setitem__({key})={value}')
        return self

    def chunk(self, *args, **kwargs):
        print(f"xr.chunk({args}, {kwargs})")
        return self

    def astype(self, *args, **kwargs):
        print(f"xr.astype({args}, {kwargs})")
        return self

    def to_zarr(self, *args, **kwargs):
        print(f"xr.to_zarr({args}, {kwargs})")
        return self

    def __getattr__(self, name):
        print(f"xr.{name} (unkwown)")
        return self


class DatasetMockup(TestingMockup):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        print(f"Climetlab SourceMockup : args={args}, kwargs={kwargs}")
        super(SourceMockup, self).__init__(**kwargs)

    def to_xarray(self, *args, **kwargs):
        return TestingDatasetAsXarray(*self.args, **self.kwargs)


class SourceMockup(Source):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        print(f"Climetlab SourceMockup : args={args}, kwargs={kwargs}")
        super(SourceMockup, self).__init__(**kwargs)

    def to_xarray(self, *args, **kwargs):
        return TestingDatasetAsXarray(*self.args, **self.kwargs)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import warnings


class GlobalIndex:
    def __init__(self, index_location, baseurl) -> None:
        warnings.warn(
            "GlobalIndex is obsolete. Please update your code and use the 'directory' source"
        )
        raise NotImplementedError(
            "GlobalIndex is obsolete. Please update your code and use the 'directory' source"
        )
        """The GloblaIndex has one index managing multiple urls/files.
        This unique index is found at "index_location"
        The path of each file is written in the index as a relative path.
        It is relative to a base url: "baseurl".
        """


class PerUrlIndex:
    def __init__(
        self,
        pattern,
    ) -> None:
        # warnings.warn( "PerUrlIndex is obsolete.")

        self.pattern = pattern



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
from climetlab.wrappers import Wrapper


class NoneWrapper(Wrapper):
    def __init__(self, data):
        pass

    def plot_map(self, backend):
        pass

    def field_metadata(self):
        return {}


def wrapper(data, *args, **kwargs):
    if data is None:
        return NoneWrapper(data)
    return None



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import json
import os
import re
import sys
from importlib import import_module

from termcolor import colored

from .tools import experimental, parse_args, print_table


def version(module):
    try:
        from importlib.metadata import PackageNotFoundError, version
    except Exception:
        from importlib_metadata import PackageNotFoundError, version

    try:
        return version(module)
    except PackageNotFoundError:
        pass

    try:
        module = import_module(module)

        if not hasattr(module, "__file__") or module.__file__ is None:
            if not hasattr(module, "__path__"):
                return "builtin"

            return "namespace"

        if hasattr(module, "__version__"):
            return str(module.__version__)

        if hasattr(module, "version"):
            return str(module.version)

        path = os.path.realpath(module.__file__)
        if os.path.basename(path) == "__init__.py":
            path = os.path.dirname(path)

        directory = os.path.basename(os.path.dirname(path))
        if directory == "lib-dynload":
            directory = os.path.basename(os.path.dirname(os.path.dirname(path)))
        if re.fullmatch(r"python\d+\.\d+", directory):
            return directory

        return module.__file__

    except ImportError:
        return "missing"
    except Exception as e:
        return e


class CheckCmd:
    @parse_args()
    @experimental
    def do_check(self, args):
        """Experimental script to help debugging."""
        import climetlab

        print(
            colored(
                (
                    "This script is experimental to help debugging."
                    "Seeing everything to 'ok' does NOT means that you have the right versions installed."
                ),
                "red",
            )
        )

        print("--------------------")

        print("Checking climetlab installation.")
        print(f"Climetlab installed in {os.path.dirname(climetlab.__file__)}")
        print("Checking required compiled dependencies...")

        import ecmwflibs

        versions = ecmwflibs.versions()

        for name in ["eccodes", "magics"]:
            try:
                print(
                    f"  {name} from ecmwflibs: ok {versions[name]} ({ecmwflibs.find(name)})"
                )
            except Exception as e:  # noqa: F841
                print(f"  {name} from ecmwflib: Warning: ecmwflibs cannot find {name}")

        for name in ["eccodes", "MagPlus", "netcdf"]:
            try:
                import findlibs
            except Exception as e:  # noqa: F841
                print(f"    {name} from findlibs: Warning: cannot import findlibs")
                continue
            try:
                print(f"    {name} from findlibs: ({findlibs.find(name)})")
            except Exception as e:  # noqa: F841
                print(f"    {name} from findlibs: Warning: findlibs cannot find {name}")

        print("Checking required python dependencies...")
        for name in ["xarray", "Magics", "eccodes", "ecmwflibs"]:
            more = ""
            try:
                lib = import_module(name)
            except ImportError:
                print(f"  Error: cannot import {name}.")
                continue
            # if name == "eccodes":
            #     more = f" (using .lib={lib.lib})"
            print(
                f"  {name}: ok {lib.__version__} ({os.path.dirname(lib.__file__)}){more}"
            )

        print("Checking optional dependencies...")
        for name in ["folium", "pdbufr", "pyodc"]:
            try:
                lib = import_module(name)
                print(
                    f"  {name}: ok {lib.__version__} ({os.path.dirname(lib.__file__)})"
                )
            except Exception as e:
                print(e)
                print(f"  Warning: cannot import {name}. Limited capabilities.")

        # TODO: add more
        # TODO: automate this from requirements.txt. Create a pip install climetlab[extra] or climetlab-light.

    @parse_args(
        json=dict(action="store_true", help="produce a JSON output"),
    )
    def do_plugins(self, args):
        """List the available plugins"""

        import entrypoints

        plugins = []
        for kind in ("source", "dataset"):
            for e in entrypoints.get_group_all(f"climetlab.{kind}s"):
                module = e.module_name.split(".")[0]
                plugins.append((kind, e.name, e.module_name, version(module)))

        if args.json:
            result = {}
            for n in plugins:
                result[n[1]] = dict(type=n[0], module=n[2], version=n[3])
            print(json.dumps(result, indent=4, sort_keys=True))
            return

        for n in sorted(plugins):
            print(n[0], n[1])
            print("  module:", n[2], "version:", n[3])
            print()

    def _loaded_modules(self, name):
        import pkgutil

        loader = pkgutil.get_loader(name)
        path = os.path.realpath(loader.path)
        if os.path.isfile(path):
            path = os.path.dirname(path)

        modules = set()
        for root, _, files in os.walk(path):
            for file in files:
                if not file.endswith(".py"):
                    continue
                full = os.path.join(root, file)
                with open(full, "rt") as f:
                    for line in f.readlines():
                        line = line.strip()
                        if line.startswith("import ") or line.startswith("from "):
                            module = line.split(" ")[1].split(".")[0]
                            if module:
                                modules.add(module)

        modules.add("dask")
        modules.add("distributed")
        modules.add("asyncssh")
        return modules

    @parse_args(
        modules=(
            "modules",
            dict(
                metavar="PACKAGE",
                type=str,
                nargs="*",
                help="optional list of Python packages",
            ),
        ),
        json=dict(action="store_true", help="produce a JSON output"),
        all=dict(action="store_true"),
    )
    def do_versions(self, args):
        """List the versions of important Python packages."""
        import entrypoints

        modules = set(args.modules)

        if not modules:
            modules = self._loaded_modules("climetlab")
            seen = set()
            for kind in ("source", "dataset"):
                for e in entrypoints.get_group_all(f"climetlab.{kind}s"):
                    name = e.module_name.split(".")[0]
                    if name not in seen:
                        modules.update(self._loaded_modules(name))
                        seen.add(name)

        result = {}

        if args.all:
            for module in modules:
                try:
                    import_module(module)
                except Exception:
                    pass
            modules.update(sys.modules.keys())

        modules = set(m.split(".")[0] for m in modules if not m.startswith("_"))

        for module in modules:
            result[module] = version(module)

        if args.json:
            print(json.dumps(result, indent=4, sort_keys=True))
        else:
            COLORS = dict(
                missing="red", damaged="red", builtin="blue", namespace="magenta"
            )
            items = []
            colours = []
            for k, v in sorted(result.items()):
                items.append((k, v))
                if not isinstance(v, str):
                    v = str(v)
                    c = "red"
                else:
                    c = "yellow" if v.startswith("python") else "green"

                colours.append(COLORS.get(v, c))

            print_table(items, colours)

    @parse_args(
        json=dict(action="store_true", help="produce a JSON output"),
    )
    def do_libraries(self, args):
        result = {}
        try:
            import Magics

            result["magics"] = os.path.realpath(Magics.lib)
        except Exception as e:
            result["magics"] = str(e)

        try:
            import gribapi

            result["eccodes"] = os.path.realpath(gribapi.library_path)
        except Exception as e:
            result["eccodes"] = str(e)

        if args.json:
            print(json.dumps(result, indent=4, sort_keys=True))
        else:
            items = []
            colours = []
            for k, v in sorted(result.items()):
                items.append((k, v))
                colours.append("green" if os.path.exists(v) else "red")

            print_table(items, colours)

    def do_df(self, path):
        from climetlab.core.caching import disk_usage

        if path == "-h":
            print("Provide some information about disk usage.")
            return

        print(disk_usage(path))



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
from contextlib import contextmanager

from climetlab import settings
from climetlab.loaders import HDF5Loader, ZarrLoader
from climetlab.utils.humanize import list_to_human

from .tools import parse_args


class LoadersCmd:
    @parse_args(
        # dataset=(
        #     "--dataset",
        #     dict(
        #         help="Name of the HDF5 dataset to use"
        #         " (default from config or 'dataset')"
        #     ),
        # ),
        target=(
            "--target",
            dict(
                help="Where to store the final data. "
                "Currently only a path to a new ZARR is supported."
            ),
        ),
        init=(
            "--init",
            dict(action="store_true", help="Initialise zarr."),
        ),
        load=(
            "--load",
            dict(action="store_true", help="Load data into zarr."),
        ),
        statistics=(
            "--statistics",
            dict(action="store_true", help="Compute statistics."),
        ),
        config=(
            "--config",
            dict(
                help="Use with --init. A yaml file that describes which data to use as input"
                " and how to organise them in the target."
            ),
        ),
        parts=(
            "--parts",
            dict(nargs="+", help="Use with --load. Part(s) of the data to process."),
        ),
        no_write=(
            "--no-write",
            dict(
                action="store_true", help="Only compute statistics, do not write them."
            ),
        ),
        cache_dir=(
            "--cache-dir",
            dict(
                help="Use with --load. Location of cache directory for temporary data."
            ),
        ),
        format=(
            "--format",
            dict(
                help="The format of the target storage into which to load the data"
                " (default is inferred from target path extension)"
                " only .zarr is currently supported."
            ),
        ),
        no_check=(
            "--no-check",
            dict(action="store_true", help="Skip checks."),
        ),
        force=(
            "--force",
            dict(action="store_true", help="Overwrite if already exists."),
        ),
        timeout=(
            "--timeout",
            dict(
                type=int,
                default=0,
                help="Stop with error (SIGALARM) after TIMEOUT seconds.",
            ),
        ),
    )
    def do_create(self, args):
        create(args)


def create(args):
    format = args.format

    if args.timeout:
        import signal

        signal.alarm(args.timeout)

    if format is None:
        _, ext = os.path.splitext(args.target)
        format = ext[1:]
    assert format == "zarr", f"Unsupported format={format}"

    def no_callback(*args, **kwargs):
        print(*args, **kwargs)
        return

    if os.environ.get("CLIMETLAB_CREATE_SHELL_CALLBACK"):

        def callback(*msg):
            msg = [str(_) for _ in msg]
            msg = "\n".join(msg)
            import shlex
            import subprocess
            import traceback

            cmd = os.environ.get("CLIMETLAB_CREATE_SHELL_CALLBACK")
            cmd = cmd.format(msg)
            try:
                print(f"Running {cmd}")
                args = shlex.split(cmd)  # shlex honors the quotes
                subprocess.run(args)
            except Exception as e:
                print(f"Exception when running {cmd}" + traceback.format_exc())
                print(e)

        callback("Starting-zarr-loader.")
    else:
        callback = no_callback

    LOADERS = dict(
        zarr=ZarrLoader,
        h5=HDF5Loader,
        hdf5=HDF5Loader,
        hdf=HDF5Loader,
    )
    if format not in LOADERS:
        lst = list_to_human(list(LOADERS.keys()), "or")
        raise ValueError(f"Invalid format '{format}', must be one of {lst}.")

    kwargs = vars(args)
    kwargs["path"] = kwargs["target"]
    kwargs["print"] = callback
    loader_class = LOADERS[format]

    lst = [args.load, args.statistics, args.init]
    if sum(1 for x in lst if x) > 1:
        raise ValueError(
            "Too many options provided."
            'Must choose exactly one option in "--load", "--statistics", "--init"'
        )
    if sum(1 for x in lst if x) < 1:
        raise ValueError(
            "Not enough options provided."
            'Must choose exactly one option in "--load", "--statistics", "--init"'
        )
    if args.parts:
        assert args.load, "Use --parts only with --load"
    if args.no_write:
        assert args.statistics, "Use --no-write only with --statistics"

    @contextmanager
    def dummy_context():
        yield

    context = dummy_context()
    if kwargs["cache_dir"]:
        context = settings.temporary("cache-directory", kwargs["cache_dir"])

    with context:
        if args.init:
            assert args.config, "--init requires --config"
            assert args.target, "--init requires --target"

            import zarr

            try:
                zarr.open(args.target, "r")
                if not args.force:
                    raise Exception(
                        f"{args.target} already exists. Use --force to overwrite."
                    )
            except zarr.errors.PathNotFoundError:
                pass

            loader = loader_class.from_config(**kwargs)
            loader.initialise()
            exit()

        if args.load:
            assert args.config is None, "--load requires only a --target, no --config."
            loader = loader_class.from_zarr(**kwargs)
            loader.load(**kwargs)

        if args.statistics:
            assert (
                args.config is None
            ), "--statistics requires only --target, no --config."
            loader = loader_class.from_zarr(**kwargs)
            loader.add_statistics(**kwargs)



# (C) Copyright 2023 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import re

LOG = logging.getLogger(__name__)


class Remapping(dict):
    # inherit from dict to make it serialisable

    def __init__(self, remapping):
        super().__init__(remapping)

        self.lists = {}
        for k, v in remapping.items():
            if isinstance(v, str):
                v = re.split(r"\{([^}]*)\}", v)
            self.lists[k] = v

    def __call__(self, func):
        if not self:
            return func

        class CustomJoiner:
            def format_name(self, x):
                return func(x)

            def format_string(self, x):
                return str(x)

            def join(self, args):
                return "".join(str(x) for x in args)

        joiner = CustomJoiner()

        def wrapped(name):
            return self.substitute(name, joiner)

        return wrapped

    def substitute(self, name, joiner):
        if name in self.lists:
            lst = []
            for i, bit in enumerate(self.lists[name]):
                if i % 2:
                    p = joiner.format_name(bit)
                    if p is not None:
                        lst.append(p)
                    else:
                        lst = lst[:-1]
                else:
                    lst.append(joiner.format_string(bit))
            return joiner.join(lst)
        return joiner.format_name(name)

    def as_dict(self):
        return dict(self)


def build_remapping(mapping):
    if mapping is None:
        return Remapping({})

    if not isinstance(mapping, Remapping) and isinstance(mapping, dict):
        return Remapping(mapping)

    return mapping


def normalize_order_by(*args, **kwargs):
    _kwargs = {}
    for a in args:
        if a is None:
            continue
        if isinstance(a, dict):
            _kwargs.update(a)
            continue
        if isinstance(a, str):
            _kwargs[a] = "ascending"
            continue
        if isinstance(a, (list, tuple)):
            if not all(isinstance(k, str) for k in a):
                _kwargs.update(normalize_order_by(*a))
            else:
                for k in a:
                    _kwargs[k] = "ascending"
            continue

        raise ValueError(f"Unsupported argument {a} of type {type(a)}")

    _kwargs.update(kwargs)

    for k, v in _kwargs.items():
        if not (
            v is None
            or callable(v)
            or isinstance(v, (list, tuple, set))
            or v in ["ascending", "descending"]
        ):
            raise ValueError(f"Unsupported order: {v} of type {type(v)} for key {k}")

    return _kwargs



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import cdsapi
import yaml

from climetlab.core.thread import SoftThreadPool
from climetlab.decorators import normalize
from climetlab.utils import tqdm

from .file import FileSource
from .prompt import APIKeyPrompt


class ADSAPIKeyPrompt(APIKeyPrompt):
    register_or_sign_in_url = "https://ads.atmosphere.copernicus.eu/"
    retrieve_api_key_url = "https://ads.atmosphere.copernicus.eu/api-how-to"

    prompts = [
        dict(
            name="url",
            default="https://ads.atmosphere.copernicus.eu/api/v2",
            title="API url",
            validate=r"http.?://.*",
        ),
        dict(
            name="key",
            example="123:abcdef01-0000-1111-2222-0123456789ab",
            title="API key",
            hidden=True,
            validate=r"\d+:[\-0-9a-f]+",
        ),
    ]

    rcfile = "~/.adsapirc"

    def save(self, input, file):
        yaml.dump(input, file, default_flow_style=False)


def client():
    prompt = ADSAPIKeyPrompt()
    prompt.check()

    path = os.path.expanduser("~/.adsapirc")

    if not os.path.exists(path):
        prompt.ask_user_and_save()

    with open(path) as f:
        rc = yaml.loads(f.read())

    return cdsapi.Client(**rc)


EXTENSIONS = {
    "grib": ".grib",
    "netcdf": ".nc",
}


class AdsRetriever(FileSource):
    sphinxdoc = """
    AdsRetriever
    """

    def __init__(self, dataset, *args, **kwargs):
        super().__init__()

        assert isinstance(dataset, str)
        if len(args):
            assert len(args) == 1
            assert isinstance(args[0], dict)
            assert not kwargs
            kwargs = args[0]

        requests = self.requests(**kwargs)

        client()  # Trigger password prompt before thraeding

        nthreads = min(self.settings("number-of-download-threads"), len(requests))

        if nthreads < 2:
            self.path = [self._retrieve(dataset, r) for r in requests]
        else:
            with SoftThreadPool(nthreads=nthreads) as pool:
                futures = [pool.submit(self._retrieve, dataset, r) for r in requests]

                iterator = (f.result() for f in futures)
                self.path = list(tqdm(iterator, leave=True, total=len(requests)))

    def _retrieve(self, dataset, request):
        def retrieve(target, args):
            client().retrieve(args[0], args[1], target)

        return self.cache_file(
            retrieve,
            (dataset, request),
            extension=EXTENSIONS.get(request.get("format"), ".cache"),
        )

    @normalize("date", "date-list(%Y-%m-%d)")
    @normalize("area", "bounding-box(list)")
    def requests(self, **kwargs):
        # TODO: move these 5 lines into @normalize
        if "year" in kwargs:
            if "month" not in kwargs:
                kwargs["month"] = [f"{i+1:02}" for i in range(0, 12)]
            if "day" not in kwargs:
                kwargs["day"] = [f"{i+1:02}" for i in range(0, 31)]

        split_on = kwargs.pop("split_on", None)
        if split_on is None or not isinstance(kwargs.get(split_on), (list, tuple)):
            return [kwargs]

        result = []

        for v in kwargs[split_on]:
            r = dict(**kwargs)
            r[split_on] = v
            result.append(r)

        return result

    def to_pandas(self, **kwargs):
        pandas_read_csv_kwargs = dict(
            comment="#",
            parse_dates=["report_timestamp"],
            skip_blank_lines=True,
            compression="zip",
        )

        pandas_read_csv_kwargs.update(kwargs.get("pandas_read_csv_kwargs", {}))

        odc_read_odb_kwargs = dict(
            # TODO
        )

        odc_read_odb_kwargs.update(kwargs.get("odc_read_odb_kwargs", {}))

        return super().to_pandas(
            pandas_read_csv_kwargs=pandas_read_csv_kwargs,
            odc_read_odb_kwargs=odc_read_odb_kwargs,
            **kwargs,
        )


source = AdsRetriever



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
import weakref
from importlib import import_module

from climetlab.core import Base
from climetlab.decorators import locked

LOG = logging.getLogger(__name__)


class ReaderMeta(type(Base), type(os.PathLike)):
    pass


class Reader(Base, os.PathLike, metaclass=ReaderMeta):
    appendable = False  # Set to True if the data can be appened to and existing file
    binary = True

    def __init__(self, source, path):
        LOG.debug("Reader for %s is %s", path, self.__class__.__name__)

        self._source = weakref.ref(source)
        self.path = path

    @property
    def source(self):
        return self._source()

    @property
    def filter(self):
        return self.source.filter

    @property
    def merger(self):
        return self.source.merger

    def mutate(self):
        # Give a chance to `directory` or `zip` to change the reader
        return self

    def mutate_source(self):
        # The source may ask if it needs to mutate
        return None

    def ignore(self):
        # Used by multi-source
        return False

    def cache_file(self, *args, **kwargs):
        return self.source.cache_file(*args, **kwargs)

    def save(self, path):
        mode = "wb" if self.binary else "w"
        with open(path, mode) as f:
            self.write(f)

    def write(self, f):
        if not self.appendable:
            assert f.tell() == 0
        mode = "rb" if self.binary else "r"
        with open(self.path, mode) as g:
            while True:
                chunk = g.read(1024 * 1024)
                if not chunk:
                    break
                f.write(chunk)

    def __fspath__(self):
        return self.path

    def index_content(self):
        LOG.warning(f"index-content(): Ignoring {self.path}")
        return []


_READERS = {}


# TODO: Add plugins
@locked
def _readers():
    if not _READERS:
        here = os.path.dirname(__file__)
        for path in sorted(os.listdir(here)):
            if path[0] in ("_", "."):
                continue

            if path.endswith(".py") or os.path.isdir(os.path.join(here, path)):
                name, _ = os.path.splitext(path)

                try:
                    module = import_module(f".{name}", package=__name__)
                    if hasattr(module, "reader"):
                        _READERS[name] = module.reader
                        if hasattr(module, "aliases"):
                            for a in module.aliases:
                                assert a not in _READERS
                                _READERS[a] = module.reader
                except Exception:
                    LOG.exception("Error loading reader %s", name)

    return _READERS


def reader(source, path):
    assert isinstance(path, str), source

    if hasattr(source, "reader"):
        reader = source.reader
        LOG.debug("Looking for a reader for %s (%s)", path, reader)
        if callable(reader):
            return reader(source, path)
        if isinstance(reader, str):
            return _readers()[reader.replace("-", "_")](source, path, None, False)

        raise TypeError(
            "Provided reader must be a callable or a string, not %s" % type(reader)
        )

    if os.path.isdir(path):
        from .directory import DirectoryReader

        return DirectoryReader(source, path).mutate()
    LOG.debug("Reader for %s", path)

    with open(path, "rb") as f:
        magic = f.read(8)

    LOG.debug("Looking for a reader for %s (%s)", path, magic)

    for deeper_check in (False, True):
        # We do two passes, the second one
        # allow the plugin to look deeper in the file
        for name, r in _readers().items():
            reader = r(source, path, magic, deeper_check)
            if reader is not None:
                return reader.mutate()

    from .unknown import Unknown

    return Unknown(source, path, magic)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging

LOG = logging.getLogger(__name__)


def _identity(x):
    return x


class ListMixin:
    def cast(self, value):
        if isinstance(value, tuple):  # We choose list over tuple
            value = list(value)
        if not isinstance(value, list):
            value = [value]

        return [self._cast(v) for v in value]

    def format(self, value, format):
        return [self._format(v, format) for v in value]


class NonListMixin:
    def cast(self, value):
        if isinstance(value, (tuple, list)):
            if len(value) != 1:
                raise TypeError(
                    f"Expected non-list but was {type(value).__name__}: {value}"
                )
            value = value[0]
        return self._cast(value)

    def format(self, value, format):
        return self._format(value, format)


class SingleOrListMixin:
    def cast(self, value):
        if isinstance(value, list):
            return [self._cast(v) for v in value]
        if isinstance(value, tuple):
            return tuple([self._cast(v) for v in value])
        return self._cast(value)

    def format(self, value, format):
        if isinstance(value, list):
            return [self._format(v, format) for v in value]
        if isinstance(value, tuple):
            return tuple([self._format(v, format) for v in value])
        return self._format(value, format)


class Type:
    def _cast(self, value):
        raise NotImplementedError(self.__class__)

    def _format(self, value, format):
        return format % (value,)

    def __repr__(self):
        return self.__class__.__name__

    def update(self, availability):
        pass


class _EnumType(Type):
    def __init__(self, values):
        self.values = values

    def _cast(self, value):
        def same(a, b):
            if a == b:
                return True
            try:
                return a.upper() == b.upper()
            except AttributeError:
                pass
            try:
                return float(a) == float(b)
            except (ValueError, TypeError):
                pass
            return False

        for v in self.values:
            if same(value, v):
                return v

        raise ValueError(
            f"Invalid value '{value}', possible values are {self.values} ({self.__class__.__name__})"
        )

    def update(self, availability):
        # TODO: if value is none : use availability w
        # else : check subset : values included into availability
        pass


class EnumSingleOrListType(_EnumType, SingleOrListMixin):
    pass


class EnumType(_EnumType, NonListMixin):
    pass


class EnumListType(_EnumType, ListMixin):
    def cast(self, value):
        from climetlab.arguments.transformers import ALL

        if value is ALL:
            return self.values
        return super().cast(value)


class _StrType(Type):
    def _cast(self, value):
        return str(value)


class StrType(_StrType, NonListMixin):
    pass


class StrListType(_StrType, ListMixin):
    pass


class StrSingleOrListType(_StrType, SingleOrListMixin):
    pass


class _AnyType(Type):
    def _cast(self, value):
        return value


class AnyType(_AnyType, NonListMixin):
    pass


class AnyListType(_AnyType, ListMixin):
    pass


class AnySingleOrListType(_AnyType, SingleOrListMixin):
    def _cast(self, value):
        return value


class _IntType(Type):
    def _cast(self, value):
        return int(value)


class IntType(_IntType, NonListMixin):
    pass


class IntListType(_IntType, ListMixin):
    def cast(self, value):
        if isinstance(value, str) and "/" in value:
            bits = value.split("/")
            if len(bits) == 3 and bits[1].lower() == "to":
                value = list(range(int(bits[0]), int(bits[2]) + 1, 1))

            elif len(bits) == 5 and bits[1].lower() == "to" and bits[3].lower() == "by":
                value = list(
                    range(int(bits[0]), int(bits[2]) + int(bits[4]), int(bits[4]))
                )
        return super().cast(value)


class IntSingleOrListType(_IntType, SingleOrListMixin):
    pass


class _FloatType(Type):
    def _cast(self, value):
        return float(value)


class FloatType(_FloatType, NonListMixin):
    pass


class FloatListType(_FloatType, ListMixin):
    pass


class FloatSingleOrListType(_FloatType, SingleOrListMixin):
    pass


class _DateType(Type):
    def _format(self, value, format):
        if format == "datetime.datetime":
            return value
        if format == datetime.datetime:
            return value
        return value.strftime(format)

    def include_args(self, decorator, args):
        assert len(args) == 1, args
        decorator.format = args[0]


class DateType(_DateType, NonListMixin):
    def cast(self, value):
        from climetlab.utils.dates import to_date_list

        # TODO: change into to_datetime?
        lst = to_date_list(value)
        assert len(lst) == 1, lst
        return lst[0]

    def _cast(self, value):
        return value


class DateListType(_DateType, ListMixin):
    def cast(self, value):
        from climetlab.utils.dates import to_date_list

        lst = to_date_list(value)
        return lst


class DateSingleOrListType(_DateType, SingleOrListMixin):
    def cast(self, value):
        from climetlab.utils.dates import to_date_list

        if isinstance(value, list):
            return to_date_list(value)
        if isinstance(value, tuple):
            return tuple(to_date_list(value))

        # TODO: change into to_datetime?
        lst = to_date_list(value)
        assert len(lst) == 1, lst
        return lst[0]

    def _cast(self, value):
        return value


class _VariableType(Type):
    def __init__(self, convention):
        self.convention = convention

    def _cast(self, value):
        from climetlab.utils.conventions import normalise_string

        return normalise_string(str(value), convention=self.convention)


class VariableType(_VariableType, NonListMixin):
    pass


class VariableListType(_VariableType, ListMixin):
    pass


class VariableSingleOrListType(_VariableType, SingleOrListMixin):
    pass


class BoundingBoxType(Type, NonListMixin):
    def include_args(self, decorator, args):
        assert len(args) == 1, args
        decorator.format = args[0]

    def cast(self, value):
        from climetlab.utils.bbox import to_bounding_box

        return to_bounding_box(value)

    def format(self, value, format):
        from climetlab.utils.bbox import BoundingBox

        FORMATTERS = {
            list: lambda x: x.as_list(),
            tuple: lambda x: x.as_tuple(),
            dict: lambda x: x.as_dict(),
            BoundingBox: _identity,
            "list": lambda x: x.as_list(),
            "tuple": lambda x: x.as_tuple(),
            "dict": lambda x: x.as_dict(),
            "BoundingBox": _identity,
            None: _identity,
        }
        formatter = FORMATTERS[format]
        return formatter(value)


GIVEN_TYPES = {
    int: "int",
    float: "float",
    str: "str",
    datetime.datetime: "date",
    bool: "bool",  # TODO: implement me
}

NON_LIST_TYPES = {
    "int": IntType,
    "float": FloatType,
    "str": StrType,
    "enum": EnumType,
    "date": DateType,
    "variable": VariableType,
    "bounding-box": BoundingBoxType,
    "bbox": BoundingBoxType,
}
LIST_TYPES = {
    "int-list": IntListType,
    "float-list": FloatListType,
    "str-list": StrListType,
    "enum-list": EnumListType,
    "date-list": DateListType,
    "variable-list": VariableListType,
}

SINGLE_OR_LIST_TYPES = {
    "int": IntSingleOrListType,
    "float": FloatSingleOrListType,
    "str": StrSingleOrListType,
    "enum": EnumSingleOrListType,
    "date": DateSingleOrListType,
    "variable": VariableSingleOrListType,
}


def infer_type(**kwargs):
    LOG.debug("INFER => %s", kwargs)
    x = _infer_type(**kwargs)
    LOG.debug("INFER <= %s", x)
    return x


def _infer_type(**kwargs):
    type = kwargs.pop("type", None)
    values = kwargs.pop("values", None)
    multiple = kwargs.pop("multiple", None)

    # TODO:
    assert not isinstance(type, Type), f"IMPLEMENT infer_type({type})"

    # Take care of builtin types and others
    if type in GIVEN_TYPES:
        return infer_type(
            type=GIVEN_TYPES[type],
            values=values,
            multiple=multiple,
            **kwargs,
        )

    # normalize("name", ["a", "b", "c"]) and similar
    if isinstance(values, (list, tuple)):  # and type is None:
        if type not in (None, "enum", "enum-list"):
            LOG.warning(
                f"Type ignored with enums, values={values}, type={type} and multiple={multiple}"
            )
        if multiple is None and type == "enum-list":
            multiple = True

        if multiple is None:
            return EnumSingleOrListType(values)
        if multiple is True:
            return EnumListType(values)
        if multiple is False:
            return EnumType(values)

    if values is None and isinstance(type, str):
        if multiple is None:
            try:
                if type in SINGLE_OR_LIST_TYPES:
                    return SINGLE_OR_LIST_TYPES[type](**kwargs)
                return {**LIST_TYPES, **NON_LIST_TYPES}[type](**kwargs)
            except Exception as e:
                raise ValueError(f"Error building {type}({kwargs}): {e}")

        if multiple is False:
            if type in LIST_TYPES:
                raise ValueError(f"Cannot set multiple={multiple} and type={type}.")
            return NON_LIST_TYPES[type](**kwargs)

        if multiple is True:
            if type in LIST_TYPES:
                return LIST_TYPES[type](**kwargs)
            if type + "-list" in LIST_TYPES:
                return LIST_TYPES[type + "-list"](**kwargs)
            raise ValueError(
                f"Cannot set multiple={multiple} and type={type}. Type must be in {list(LIST_TYPES.keys())}"
            )

    if values is None and type is None:
        if multiple is True:
            return AnyListType()
        if multiple is False:
            return AnyType()
        if multiple is None:
            return AnySingleOrListType()

    raise ValueError(
        f"Cannot infer type from values={values}, type={type} and multiple={multiple}"
    )



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import ecmwf.opendata

from .file import FileSource


class EODRetriever(FileSource):
    sphinxdoc = """
    EODRetriever
    """

    def __init__(self, source="ecmwf", *args, **kwargs):
        if len(args):
            assert len(args) == 1
            assert isinstance(args[0], dict)
            assert not kwargs
            kwargs = args[0]

        self.source_kwargs = self.request(**kwargs)

        self.client = ecmwf.opendata.Client(source=source, preserve_request_order=True)

        self.path = self._retrieve(self.source_kwargs)

    def connect_to_mirror(self, mirror):
        return mirror.connection_for_eod(self)

    def _retrieve(self, request):
        def retrieve(target, request):
            self.client.retrieve(request, target)

        return self.cache_file(
            retrieve,
            request,
        )

    # @normalize("date", "date-list(%Y-%m-%d)")
    # @normalize("area", "bounding-box(list)")
    def request(self, **request):
        return request


source = EODRetriever



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from collections import defaultdict
from functools import partial

from climetlab.core.data import data_entries, get_data_entry
from climetlab.core.ipython import Image, display
from climetlab.core.settings import SETTINGS
from climetlab.core.temporary import temp_file
from climetlab.wrappers import get_wrapper

from .options import Options

LOG = logging.getLogger(__name__)

OPTIONS = {}


def magics(*args, **kwargs):
    from .backends.magics.backend import Backend as MagicsBackend

    return MagicsBackend(*args, **kwargs)


def options(**kwargs):
    global OPTIONS
    OPTIONS = kwargs


def projection(name):
    return get_data_entry("projections", name)


def layer(name):
    return get_data_entry("layers", name)


def style(name):
    return get_data_entry("styles", name)


def projections():
    return sorted(e.name for e in data_entries("projections"))


def layers():
    return sorted(e.name for e in data_entries("layers"))


def styles():
    return sorted(e.name for e in data_entries("styles"))


class Plot:
    """[summary]"""

    def __init__(self, kwargs):
        if "backend" in kwargs:
            backend = kwargs.pop("backend")
            LOG.warning(f"Ignoring argument backend='{backend}'")
        backend = magics

        options = {}
        options.update(SETTINGS.get("plotting-options", {}))
        options.update(OPTIONS)
        options.update(kwargs)
        self.backend = backend(Options(options))

    def plot_map(self, data=None, metadata=None, **kwargs):
        if not isinstance(data, (list, tuple)):
            data = [data]

        if metadata:
            self.backend._options.push("metadata", metadata)

        for d in data:
            d = get_wrapper(d)
            d.plot_map(self.backend)

        options = Options(kwargs)
        self.backend.apply_options(options)
        options.check_unused()

        return self

    def wms_layers(self):
        return self.backend.wms_layers()

    def render(self):
        return self.backend.render()

    def show(self):
        self.backend.show(display=display)

    def macro(self) -> list:
        return self.backend.macro()

    def save(self, path):
        return self.backend.save(path)


def files_to_apng(files, path, fps):
    import imageio

    try:
        from numpngw import write_apng
    except ImportError:
        raise RuntimeError(
            "Package numpngw is required to save animated PNGs (pip install numpngw)"
        )

    frames = [imageio.imread(f) for f in files]

    write_apng(path, frames, delay=int(1000.0 / fps + 0.5))

    return path


def files_to_gif(files, path, fps):
    import imageio

    with imageio.get_writer(path, mode="I", duration=1.0 / fps) as writer:
        for filename in files:
            image = imageio.imread(filename)
            writer.append_data(image)

    return path


def files_to_opencv(files, path, fps, cccc):
    try:
        import cv2
    except ImportError:
        raise RuntimeError(
            "Package OpenCV is required to save movies (pip install opencv-python)"
        )

    frame = cv2.imread(files[0])
    height, width, _ = frame.shape

    video = cv2.VideoWriter(
        path,
        cv2.VideoWriter_fourcc(*cccc),
        fps,
        (width, height),
    )

    for image in files:
        video.write(cv2.imread(image))

    cv2.destroyAllWindows()
    video.release()


CODECS = {
    ".gif": files_to_gif,
    ".png": files_to_apng,
    ".avi": partial(files_to_opencv, cccc="XVID"),
    ".mp4": partial(files_to_opencv, cccc="mp4v"),
    ".wmv": partial(files_to_opencv, cccc="WMV2"),
    ".mov": partial(files_to_opencv, cccc="mp4v"),
}


def unsupported(files, path, fps):
    raise NotImplementedError(
        f"Unsupported format for '{path}'."
        f" Supported formats are {list(CODECS.keys())}"
    )


def files_to_movie(files, path, fps):
    _, ext = os.path.splitext(path)

    return CODECS.get(ext, unsupported)(files, path, fps)


class MapPlot(Plot):
    pass


class AnimationPlot:
    def __init__(self, fps=5, **kwargs):
        self.fps = fps
        self.kwargs = kwargs
        self.files = []

    def plot_map(self, data, **kwargs):
        self.files.append(temp_file(".png"))

        forward_kwargs = {}
        for forward in ("row", "column"):
            if forward in kwargs:
                forward_kwargs[forward] = kwargs.pop(forward)

        options = {}
        options.update(self.kwargs)
        options.update(kwargs)

        p = new_plot(**options)
        p.plot_map(data, **forward_kwargs)
        p.save(self.files[-1].path)

    def show(self):
        tmp = temp_file(".png")
        self.save(tmp.path)
        return display(Image(tmp.path))

    def save(self, path):
        return files_to_movie([f.path for f in self.files], path, self.fps)


class LayoutPlot:
    def __init__(self, rows, columns, animate=False, fps=5, **kwargs):
        self.rows = rows
        self.columns = columns
        self.padding = kwargs.pop("padding", 0)
        self.kwargs = kwargs
        self.files = defaultdict(dict)
        self.animate = animate
        self.fps = fps

    def plot_map(self, data, row, column, step=None, **kwargs):
        tmp = temp_file(".png")

        assert row >= 0 and row < self.rows, (row, self.rows)
        assert column >= 0 and column < self.columns, (column, self.columns)

        self.files[step][(row, column)] = tmp

        options = {}
        options.update(self.kwargs)
        options.update(kwargs)

        p = new_plot(**options)
        p.plot_map(data)
        p.save(tmp.path)

    def show(self):
        tmp = temp_file(".png")
        self.save(tmp.path)
        return display(Image(tmp.path))

    def save(self, path):
        if self.animate:
            files = []
            for step in sorted(self.files.keys()):
                files.append(temp_file(".png"))
                self.render(step, files[-1].path)

            files_to_movie([f.path for f in files], path, self.fps)

        else:
            self.render(None, path)

        return path

    def render(self, step, path):
        try:
            import imageio
        except ImportError:
            raise RuntimeError(
                "Package imageio is required to create animations (pip install imageio)"
            )
        import numpy as np

        WHITE = {1: 255}

        cells = {k: imageio.imread(v.path) for k, v in self.files[step].items()}
        first = cells[list(cells.keys())[0]]
        heigth, width, depth = first.shape

        image = (
            np.ones(
                (
                    heigth * self.rows + self.padding * (self.rows - 1),
                    width * self.columns + self.padding * (self.columns - 1),
                    depth,
                ),
                dtype=first.dtype,
            )
            * WHITE[first.dtype.itemsize]
        )

        for row in range(self.rows):
            for col in range(self.columns):
                cell = (row, col)
                if cell in cells:
                    row_padding = self.padding * row
                    col_padding = self.padding * col
                    image[
                        heigth * row + row_padding : heigth * (row + 1) + row_padding,
                        width * col + col_padding : width * (col + 1) + col_padding,
                        :,
                    ] = cells[cell]

        imageio.imwrite(path, image)

        return path


def new_plot(**kwargs) -> Plot:
    """[summary]

    :return: [description]
    :rtype: Plot
    """

    if "rows" in kwargs and "columns" in kwargs:
        return LayoutPlot(**kwargs)

    if kwargs.pop("animate", False):
        return AnimationPlot(**kwargs)

    return MapPlot(kwargs)


def plot_map(data=None, **kwargs):
    """Plot any data on a map.

    Args:
        data ([any]): [description]
    """

    p = new_plot(**kwargs)
    p.plot_map(data)
    p.show()


Plot.plot_map.__doc__ = plot_map.__doc__


def interactive_map(data=None, **kwargs):
    from climetlab.plotting.wms import interactive_map as wms_map

    return wms_map(data, **kwargs)


def new_table(*args, **kwargs):
    from climetlab.notebook.table import Table

    return Table(*args, **kwargs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

# Keep linters happy
# N801 = classes should start with uppercase
# N806 = variables should be lower case

import logging

from .convertions import convert

LOG = logging.getLogger(__name__)


class NoMagics:
    def plot(self, *args, **kwargs):
        raise NotImplementedError(
            "Magics was not loaded successfully, plotting is not supported."
        )


try:
    import Magics
    from Magics import macro

    try:
        Magics.strict_mode()
    except Exception as e:
        LOG.warning(str(e))
except Exception as e:
    LOG.warning(str(e))
    macro = NoMagics()


LOG = logging.getLogger(__name__)


class Action:
    default_style: object = None

    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __repr__(self):
        x = ["macro.%s(" % (self.action,)]
        for k, v in sorted(self.kwargs.items()):
            x.append("\n   %s=%r," % (k, v))
        x.append("\n    )")
        return "".join(x)

    def to_yaml(self):
        return {self.action: self.kwargs}

    @property
    def action(self):
        return self.__class__.__name__

    def execute(self):
        return getattr(macro, self.action)(
            **convert(self.action, self.kwargs)
        ).execute()

    def update(self, action, values):
        if not isinstance(self, action):
            return None
        for k, v in values.items():
            if k[0] in ("+",):
                self.kwargs[k[1:]] = v
            if k[0] in ("-",):
                self.kwargs.pop(k[1:], None)
            if k[0] in ("=",) and k[1:] not in self.kwargs:
                self.kwargs[k[1:]] = v
        return self


class mcont(Action):  # noqa: N801
    pass


class mcoast(Action):  # noqa: N801
    pass


class mmap(Action):  # noqa: N801
    def page_ratio(self):
        # TODO: Use projection
        south = self.kwargs.get("subpage_lower_left_latitude", -90.0)
        west = self.kwargs.get("subpage_lower_left_longitude", -180)
        north = self.kwargs.get("subpage_upper_right_latitude", 90.0)
        east = self.kwargs.get("subpage_upper_right_longitude", 180.0)
        return (north - south) / (east - west)


class FieldAction(Action):
    default_style = mcont(contour_automatic_setting="ecmwf", legend=False)


class mgrib(FieldAction):  # noqa: N801
    pass


class mnetcdf(FieldAction):  # noqa: N801
    pass


class minput(FieldAction):  # noqa: N801
    pass


class mtable(Action):  # noqa: N801
    pass


class mtext(Action):  # noqa: N801
    pass


class msymb(Action):  # noqa: N801
    pass


class output(Action):  # noqa: N801
    pass


def plot(*args, **kwargs):
    return macro.plot(*args, **kwargs)


def lookup(name):
    return globals()[name]



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


def _concat(sources, *args, **kwargs):
    ds = sources[0].to_tfdataset(**kwargs)
    for s in sources[1:]:
        ds = ds.concatenate(s.to_tfdataset(**kwargs))
    return ds


def _zip(sources, *args, **kwargs):
    import tensorflow as tf

    return tf.data.Dataset.zip(tuple(s.to_tfdataset(**kwargs) for s in sources))


def merge(
    sources=None,
    paths=None,
    reader_class=None,
    **kwargs,
):
    if paths is not None:
        if reader_class is not None and hasattr(reader_class, "to_tfdataset_multi"):
            return reader_class.to_tfdataset_multi(paths, **kwargs)

    method = "_" + kwargs.get("method", "concat")
    return globals()[method](sources, paths, reader_class, **kwargs)



#!/usr/bin/env python3# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

__version__ = "0.1.0"

import json
import logging

from climetlab.sources.url import Url
from climetlab.utils import get_json

LOG = logging.getLogger(__name__)
URLPATTERN = "https://zenodo.org/api/records/{record_id}"


class Zenodo(Url):
    def __init__(
        self,
        record_id,
        file_key=None,
        filter=None,
        merger=None,
        *args,
        **kwargs,
    ):
        record = get_json(URLPATTERN.format(record_id=record_id))
        LOG.debug("ZENODO record %s", json.dumps(record, indent=4, sort_keys=True))

        urls = {}
        for file in record["files"]:
            urls[file["key"]] = file["links"]["self"]

        if file_key is None:
            if len(urls) != 1:
                raise ValueError(
                    f"No `file_key` given, please specify on of {sorted(urls.keys())}"
                )
            file_key = list(urls.keys())[0]

        if file_key not in urls:
            raise ValueError(
                f"Invalid zenodo key '{file_key}', values are {sorted(urls.keys())}"
            )

        LOG.debug("ZENODO record_keys %s", sorted(urls.keys()))

        LOG.debug("ZENODO url %s", urls[file_key])
        super().__init__(
            urls[file_key],
            filter=filter,
            merger=merger,
            *args,
            **kwargs,
        )


source = Zenodo



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

from .tools import parse_args


class CompletionCmd:
    @parse_args(
        shell=(
            None,
            dict(
                metavar="SHELL",
                help="Shell to use for autocompletion. Must be zsh or bash.",
                nargs="?",
            ),
        ),
    )
    def do_completion(self, args):
        """
        Enable autocompletion for the "climetlab" shell command.

        Supported shells are : zsh bash.

        Usage:
            climetlab completion
        """
        from climetlab.scripts.main import command_list

        if not args.shell:
            return self.do_completion("--help")

        home = os.path.expanduser("~")

        def add_to_file(dotfile, lines):
            with open(dotfile, "a") as f:
                print(f"The following was added at the end of {dotfile} :\n")
                print("\n", file=f)
                for line in lines:
                    print(line, file=f)
                    print("  " + line)
                print("\nCompletion will be enabled for any new shell.")

        if args.shell == "zsh":
            # TODO: only add "fpath+= ( $HOME/.climetlab/autocomplete/zsh)"
            # and create the file $HOME/.climetlab/autocomplete/zsh/_climetlab)"
            dotfile = os.path.join(home, ".zshrc")
            lines = ["complete -W '" + " ".join(command_list()) + "' climetlab"]
            add_to_file(dotfile, lines)
            return

        if args.shell == "bash":
            dotfile = os.path.join(home, ".bashrc")
            lines = ["complete -W '" + " ".join(command_list()) + "' climetlab"]
            add_to_file(dotfile, lines)
            return

        raise NotImplementedError(f"Shell not supported: shell={args.shell}.")



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

from climetlab.vocabularies import VOCABULARIES

LOG = logging.getLogger(__name__)


def normalise_string(key, convention="cf"):
    vocabulary = VOCABULARIES[convention]

    new = vocabulary.normalise(key)
    LOG.debug(f"Normalising '{key}' into '{new}' ({convention} convention)")
    return new



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import itertools
import os
import zipfile

from climetlab.core.temporary import temp_directory

from .file import FileSource


def zip_dir(target, directory):
    with zipfile.ZipFile(target, "w") as z:
        for root, _, files in os.walk(directory):
            for f in files:
                fullpath = os.path.join(root, f)
                relpath = os.path.relpath(fullpath, start=directory)
                z.write(
                    fullpath,
                    arcname=relpath,
                    compress_type=zipfile.ZIP_DEFLATED,
                )


def iterate_request(r):
    yield from (dict(zip(r.keys(), x)) for x in itertools.product(*r.values()))


def generate_grib(target, **kwargs):
    import eccodes

    for k, v in list(kwargs.items()):
        if not isinstance(v, (list, tuple)):
            kwargs[k] = [v]

    handle = None
    try:
        with open(os.path.join(os.path.dirname(__file__), "dummy.grib"), "rb") as f:
            handle = eccodes.codes_new_from_file(f, eccodes.CODES_PRODUCT_GRIB)

        with open(target, "wb") as f:
            for r in iterate_request(kwargs):
                for k, v in r.items():
                    eccodes.codes_set(handle, k, v)
                eccodes.codes_write(handle, f)

    finally:
        if handle is not None:
            eccodes.codes_release(handle)


def generate_unknown(target, **kwargs):
    import json

    with open(target, "w") as f:
        print(json.dumps(kwargs), file=f)


def generate_zeros(target, size=1024 * 1024, chunk_size=1024 * 1024, **kwargs):
    chunk_size = min(chunk_size, size)
    zeros = bytes(chunk_size)

    with open(target, "wb") as f:
        while size > 0:
            bufsize = min(size, chunk_size)
            f.write(zeros[:bufsize])
            size -= bufsize


def make_xarray(
    variables=["a"],
    dims=["lat", "lon"],
    size=2,
    coord_values=None,
    attributes=None,
    **kwargs,
):
    import numpy as np
    import xarray as xr

    if isinstance(dims, (list, tuple)):
        dims = {k: dict(size=size) for k in dims}

    all_dims = list(dims.keys())

    if isinstance(variables, (list, tuple)):
        variables = {k: dict(dims=all_dims) for k in variables}

    coords = {}
    for d, args in dims.items():
        _size = args["size"]
        coords[d] = np.arange(_size, dtype=float)
        if coord_values and d in coord_values:
            coords[d] = coord_values[d]

    vars = {}
    for v, args in variables.items():
        _dims = variables[v]["dims"]
        _coords = {k: coords[k] for k in _dims}
        _shape = [dims[d]["size"] for d in _dims]
        data = np.arange(np.prod(_shape)).reshape(_shape)  # .as_type(float)
        arr = xr.DataArray(
            data,
            dims=_dims,
            coords=_coords,
        )
        vars[v] = arr

    ds = xr.Dataset(vars)

    if "lat" in ds.dims:
        ds["lat"].attrs["standard_name"] = "latitude"
    if "lon" in ds.dims:
        ds["lon"].attrs["standard_name"] = "longitude"
    if "time" in ds.dims:
        ds["time"].attrs["standard_name"] = "time"

    if attributes:
        for d, attrs in attributes.items():
            for k, v in attrs.items():
                ds[d].attrs[k] = v

    return ds


def generate_zarr(target, **kwargs):
    ds = make_xarray(**kwargs)
    ds.to_zarr(target)


def generate_zarr_zip(target, **kwargs):
    ds = make_xarray(**kwargs)
    with temp_directory() as tmpdir:
        ds.to_zarr(tmpdir)
        zip_dir(target, tmpdir)


def generate_netcdf(target, **kwargs):
    ds = make_xarray(**kwargs)
    ds.to_netcdf(target)


def generate_csv(
    target,
    headers=None,
    lines=[],
    separator=",",
    quote_strings=None,
    none_are_empty=True,
    **kwargs,
):
    assert none_are_empty
    with open(target, "w") as f:
        if headers:
            print(separator.join(headers), file=f)

        def str_or_none(x):
            if x is None:
                return ""
            return str(x)

        def repr_or_none(x):
            if x is None:
                return ""
            return repr(x)

        _ = repr_or_none if quote_strings else str_or_none

        for line in lines:
            print(separator.join(_(x) for x in line), file=f)


def generate_zip(target, sources=None, names=None, directory=None, **kwargs):
    if sources is not None:
        assert directory is None
        assert names is not None
        assert len(sources) == len(names)
        with temp_directory() as tmpdir:
            for s, n in zip(sources, names):
                s.save(os.path.join(tmpdir, n))
            zip_dir(target, tmpdir)
            return

    if directory is not None:
        assert sources is None
        assert names is None
        zip_dir(target, directory)
        return

    assert False


GENERATORS = {
    "csv": (generate_csv, ".csv"),
    "grib": (generate_grib, ".grib"),
    "netcdf": (generate_netcdf, ".nc"),
    "unknown": (generate_unknown, ".unknown"),
    "zarr-zip": (generate_zarr_zip, ".zarr.zip"),
    "zarr": (generate_zarr, ".zarr"),
    "zeros": (generate_zeros, ".zeros"),
    "zip": (generate_zip, ".zip"),
}


class ClimetlabTestingSource(FileSource):
    def __init__(self, kind, request=None, force=False, extension=None, **kwargs):
        super().__init__()

        if request is None:
            request = {}
        request.update(kwargs)

        generate, ext = GENERATORS[kind]
        if extension is not None:
            ext = extension

        def _generate(target, args):
            return generate(target, **args)

        self.path = self.cache_file(
            _generate,
            request,
            hash_extra=kind,
            extension=ext,
            force=force,
        )


source = ClimetlabTestingSource



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import itertools
import logging
from warnings import warn

import numpy as np

from climetlab.decorators import cached_method, normalize
from climetlab.indexing.cube import index_to_coords
from climetlab.readers.grib.index import FieldSet
from climetlab.utils.dates import to_datetime

LOG = logging.getLogger(__name__)


class ConstantMaker:
    def __init__(self, field):
        self.field = field
        self.shape = self.field.shape

    @cached_method
    def grid_points(self):
        return self.field.grid_points()

    @cached_method
    def ecef_xyz(self):
        # https://en.wikipedia.org/wiki/Geographic_coordinate_conversion#From_geodetic_to_ECEF_coordinates
        # We assume that the Earth is a sphere of radius 1 so N(phi) = 1
        # We assume h = 0

        lat, lon = self.grid_points()

        phi = np.deg2rad(lat)
        lda = np.deg2rad(lon)

        cos_phi = np.cos(phi)
        cos_lda = np.cos(lda)
        sin_phi = np.sin(phi)
        sin_lda = np.sin(lda)

        x = cos_phi * cos_lda
        y = cos_phi * sin_lda
        z = sin_phi

        return x, y, z

    @cached_method
    def latitude_(self):
        return self.grid_points()[0]

    def latitude(self, date):
        return self.latitude_()

    @cached_method
    def cos_latitude_(self):
        return np.cos(np.deg2rad(self.grid_points()[0]))

    def cos_latitude(self, date):
        return self.cos_latitude_()

    @cached_method
    def sin_latitude_(self):
        return np.sin(np.deg2rad(self.grid_points()[0]))

    def sin_latitude(self, date):
        return self.sin_latitude_()

    @cached_method
    def longitude_(self):
        return self.grid_points()[1]

    def longitude(self, date):
        return self.longitude_()

    @cached_method
    def cos_longitude_(self):
        return np.cos(np.deg2rad(self.grid_points()[1]))

    def cos_longitude(self, date):
        return self.cos_longitude_()

    @cached_method
    def sin_longitude_(self):
        return np.sin(np.deg2rad(self.grid_points()[1]))

    def sin_longitude(self, date):
        return self.sin_longitude_()

    def ecef_x(self, date):
        return self.ecef_xyz()[0]

    def ecef_y(self, date):
        return self.ecef_xyz()[1]

    def ecef_z(self, date):
        return self.ecef_xyz()[2]

    def julian_day(self, date):
        date = to_datetime(date)
        delta = date - datetime.datetime(date.year, 1, 1)
        julian_day = delta.days + delta.seconds / 86400.0
        return np.full((np.prod(self.field.shape),), julian_day)

    def cos_julian_day(self, date):
        radians = self.julian_day(date) / 365.25 * np.pi * 2
        return np.cos(radians)

    def sin_julian_day(self, date):
        radians = self.julian_day(date) / 365.25 * np.pi * 2
        return np.sin(radians)

    def local_time(self, date):
        lon = self.longitude(date)
        date = to_datetime(date)
        delta = date - datetime.datetime(date.year, date.month, date.day)
        hours_since_midnight = (delta.days + delta.seconds / 86400.0) * 24
        return (lon / 360.0 * 24.0 + hours_since_midnight) % 24

    def cos_local_time(self, date):
        radians = self.local_time(date) / 24 * np.pi * 2
        return np.cos(radians)

    def sin_local_time(self, date):
        radians = self.local_time(date) / 24 * np.pi * 2
        return np.sin(radians)

    def insolation(self, date):
        warn(
            "The function `insolation` is deprecated, please use `cos_solar_zenith_angle` instead"
        )
        return self.cos_solar_zenith_angle(date)

    def toa_incident_solar_radiation(self, date):
        from earthkit.meteo.solar import toa_incident_solar_radiation

        date = to_datetime(date)
        result = toa_incident_solar_radiation(
            date - datetime.timedelta(minutes=30),
            date + datetime.timedelta(minutes=30),
            self.latitude_(),
            self.longitude_(),
            intervals_per_hour=2,
        )
        return result.flatten()

    def cos_solar_zenith_angle(self, date):
        from earthkit.meteo.solar import cos_solar_zenith_angle

        date = to_datetime(date)
        result = cos_solar_zenith_angle(
            date,
            self.latitude_(),
            self.longitude_(),
        )
        return result.flatten()


class ConstantField:
    def __init__(self, date, param, proc, shape):
        self.date = date
        self.param = param
        self.proc = proc
        self.shape = shape
        self._metadata = dict(
            valid_datetime=date if isinstance(date, str) else date.isoformat(),
            param=param,
            level=None,
            levelist=None,
            number=0,
        )

    def to_numpy(self, reshape=True, dtype=None):
        values = self.proc(self.date)
        if reshape:
            values = values.reshape(self.shape)
        if dtype is not None:
            values = values.astype(dtype)

        # assert len(values.shape) == 1, (self, self.name, reshape, dtype, values.shape)
        return values

    def __repr__(self):
        return "ConstantField(%s,%s)" % (
            self.param,
            self.date,
        )

    def metadata(self, name):
        return self._metadata[name]


def make_datetime(date, time):
    if time is None:
        return date
    if date.hour or date.minute:
        raise ValueError(
            f"Duplicate information about time time={time}, and time={date.hour}:{date.minute} from date={date}"
        )
    assert date.hour == 0, (date, time)
    assert date.minute == 0, (date, time)
    assert str(time).isdigit(), (type(time), time)
    time = int(time)
    if time > 24:
        time = time // 100
    return datetime.datetime(date.year, date.month, date.day, time)


class Constants(FieldSet):
    def __init__(self, source_or_dataset, request={}, repeat=1, **kwargs):
        request = dict(**request)
        request.update(kwargs)

        self.request = self._request(**request)

        def find_dates(request):
            if "date" not in request and "time" not in request:
                assert hasattr(
                    source_or_dataset, "unique_values"
                ), f"{source_or_dataset} (type '{type(source_or_dataset).__name__}') is not a proper source or dataset"

                return source_or_dataset.unique_values("valid_datetime")[
                    "valid_datetime"
                ]

            if "date" not in request and "time" in request:
                raise ValueError("Cannot specify time without date")

            if "date" in request and "time" not in request:
                return request["date"]

            if "date" in request and "time" in request:
                dates = [
                    make_datetime(date, time)
                    for date, time in itertools.product(
                        request["date"], request["time"]
                    )
                ]
                assert len(set(dates)) == len(dates), "Duplicates dates in constants."
                return dates

            assert False, request

        self.dates = find_dates(self.request)

        self.params = self.request["param"]
        if not isinstance(self.params, list):
            self.params = [self.params]
        self.repeat = repeat  # For ensembles
        self.maker = ConstantMaker(field=source_or_dataset[0])
        self.procs = {param: getattr(self.maker, param) for param in self.params}
        self._len = len(self.dates) * len(self.params) * self.repeat

    @normalize("date", "date-list")
    @normalize("time", "int-list")
    @normalize("number", "int-list")
    def _request(self, **request):
        return request

    def __len__(self):
        return self._len

    def _getitem(self, i):
        if i >= self._len:
            raise IndexError(i)

        date, param, repeat = index_to_coords(
            i, (len(self.dates), len(self.params), self.repeat)
        )

        assert repeat == 0, "Not implemented"

        date = self.dates[date]
        # assert isinstance(date, datetime.datetime), (date, type(date))

        param = self.params[param]
        return ConstantField(
            date,
            param,
            self.procs[param],
            self.maker.shape,
        )


source = Constants



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import numpy as np

from . import Reader


class NumpyReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)

    def to_numpy(self, numpy_load_kwargs={}):
        return np.load(self.path, **numpy_load_kwargs)

    def __iter__(self):
        return iter([self])


class NumpyZipReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)

    def to_numpy(self, numpy_load_kwargs={}):
        return np.load(self.path, **numpy_load_kwargs)


def reader(source, path, magic=None, deeper_check=False):
    if magic is None:  # Bypass check and force
        return NumpyReader(source, path)

    if magic[:6] == b"\x93NUMPY":
        return NumpyReader(source, path)

    _, extension = os.path.splitext(path)
    if magic[:4] == b"PK\x03\x04" and extension == ".npz":
        return NumpyZipReader(source, path)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging

from climetlab.arguments.climetlab_types import Type
from climetlab.vocabularies.aliases import unalias

LOG = logging.getLogger(__name__)


ALL = all


class Action:
    def execute(self, kwargs):
        raise NotImplementedError()

    def execute_before_default(self, kwargs):
        return kwargs

    def __repr__(self) -> str:
        return f"{self.__class__}"


class ArgumentTransformer(Action):
    def __init__(self, owner):
        self.owner = owner

    def execute(self, kwargs):
        if self.name in kwargs:  # TODO: discuss that
            kwargs[self.name] = self.transform(kwargs[self.name])
        return kwargs

    def transform(self, value):
        raise NotImplementedError(self.__class__.__name__)

    @property
    def name(self):
        if self.owner is None:
            return "-"
        if isinstance(self.owner, str):
            return self.owner
        return self.owner.name


class _TypedTransformer(ArgumentTransformer):
    def __init__(self, owner, type) -> None:
        super().__init__(owner)
        self.type = type if isinstance(type, Type) else type()


class AliasTransformer(_TypedTransformer):
    def __init__(self, owner, type, aliases) -> None:
        super().__init__(owner, type)
        self.aliases = aliases

        if isinstance(self.aliases, str):
            self.unalias = self.from_string
            return

        if isinstance(self.aliases, dict):
            self.unalias = self.from_dict
            return

        if callable(self.aliases):
            self.unalias = self.aliases
            return

        self.unalias = self.unsupported

    def unsupported(self, value):
        raise NotImplementedError(self.aliases)

    def from_string(self, value):
        return unalias(self.aliases, value)

    def from_dict(self, value):
        try:
            return self.aliases[value]
        except KeyError:  # No alias for this value
            pass
        except TypeError:  # if value is not hashable
            pass

        return value

    def _transform_one(self, value):
        old = object()
        while old != value:
            old = value
            value = self.unalias(old)
            LOG.debug("    Unalias %s --> %s", old, value)
        return value

    def transform(self, value):
        LOG.debug("    Unaliasing %s", value)
        if isinstance(value, list):
            return [self._transform_one(v) for v in value]
        if isinstance(value, tuple):
            return tuple([self._transform_one(v) for v in value])
        return self._transform_one(value)

    def __repr__(self) -> str:
        return f"AliasTransformer({self.owner},{self.aliases},{self.type})"


class FormatTransformer(_TypedTransformer):
    def __init__(self, owner, format, type) -> None:
        super().__init__(owner, type)
        self.format = format

    def transform(self, value):
        if value is None:
            return value
        return self.type.format(value, self.format)

    def __repr__(self) -> str:
        return f"FormatTransformer({self.owner},{self.format},{self.type})"


class TypeTransformer(_TypedTransformer):
    def __init__(self, owner, type):
        super().__init__(owner, type)

    def transform(self, value):
        if value is None:
            return value
        return self.type.cast(value)

    def __repr__(self) -> str:
        return f"TypeTransformer({self.owner},{self.type}"


class AvailabilityChecker(Action):
    def __init__(self, availability) -> None:
        self.availability = availability

    def execute(self, kwargs):
        LOG.debug("Checking availability for %s", kwargs)
        assert isinstance(kwargs, dict), kwargs
        without_none = {k: v for k, v in kwargs.items() if v is not None}
        self.availability.check(without_none)
        return kwargs

    def __repr__(self) -> str:
        txt = "Availability:"
        for line in self.availability.tree().split("\n"):
            if line:
                txt += "\n    " + line
        return txt


class KwargsAliasTransformer(Action):
    def __init__(self, alias_argument) -> None:
        self.aliases = alias_argument.kwargs

    def execute(self, kwargs):
        return kwargs

    def execute_before_default(self, kwargs):
        if kwargs:
            LOG.debug("Transforming kwargs names with aliases for %s", kwargs)
        assert isinstance(kwargs, dict), kwargs
        new_kwargs = {}
        for k, v in kwargs.items():
            new_k = self.reversed_aliases.get(k, k)
            assert (
                new_k not in new_kwargs
            ), f"Error: Multiple values were given for aliased arguments: with '{k}' and '{new_k}'."
            new_kwargs[new_k] = v
        return new_kwargs

    @property
    def reversed_aliases(self):
        reversed = {}
        for target, aliases in self.aliases.items():
            for alias in aliases:
                assert alias not in reversed, (
                    "Error: Multiple target value for alias "
                    f" argument '{alias}': '{target}' and '{reversed[alias]}'"
                )
                reversed[alias] = target
        return reversed

    def __repr__(self) -> str:
        txt = "KwargsAlias:"
        txt += ",".join([f"{v}->{k}" for k, v in self.aliases.items()])
        return txt
